from __future__ import annotations

from grim.assets import TextureId
from grim.audio import AudioState
from grim.config import CrimsonConfig
from grim.console import ConsoleState
from grim.geom import Vec2
from grim.math import clamp
from grim.rand import Crand
from grim.raylib_api import rl
from grim.view import ViewContext

from ..game_modes import GameMode
from ..input_codes import input_code_is_down, input_code_is_pressed
from ..perks.selection import perk_selection_prepared_choices
from ..persistence.save_status import GameStatusData
from ..replay import ReplayHeader, ReplayRecorder
from ..replay.checkpoints import DEFAULT_CHECKPOINT_SAMPLE_RATE
from ..sim.bootstrap import advance_unlock_terrain
from ..sim.input import PlayerInput
from ..sim.session_builders import build_tutorial_session
from ..sim.sessions import DeterministicSession
from ..ui.cursor import draw_menu_cursor
from ..ui.hud import HudRenderContext, draw_hud_overlay, hud_flags_for_game_mode
from ..ui.overlays.tutorial_run import (
    TUTORIAL_PANEL_POS,
    draw_tutorial_overlay_panels,
    tutorial_prompt_panel_rect,
)
from ..ui.perk_menu import (
    UiButtonState,
    button_draw,
    button_update,
    button_width,
)
from ..weapon_runtime import weapon_assign_player
from ..weapons import WeaponId
from .base_gameplay_mode import BaseGameplayMode
from .components.perk_menu_controller import PerkMenuController

UI_HINT_COLOR = rl.Color(140, 140, 140, 255)


class TutorialMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        demo_mode_active: bool = False,
        config: CrimsonConfig,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: Crand,
    ) -> None:
        super().__init__(
            ctx,
            world_size=1024.0,
            default_game_mode_id=GameMode.TUTORIAL,
            demo_mode_active=bool(demo_mode_active),
            quest_fail_retry_count=0,
            hardcore=False,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._perk_menu = PerkMenuController()

        self._skip_button = UiButtonState("Skip tutorial", force_wide=True)
        self._play_button = UiButtonState("Play a game", force_wide=True)
        self._repeat_button = UiButtonState("Repeat tutorial", force_wide=True)
        self._sim_session: DeterministicSession | None = None
        self._replay_recorder: ReplayRecorder | None = None
        self._frame_input_state: PlayerInput | None = None
        self._perk_pick_pending = False

    def _new_sim_session(self) -> DeterministicSession:
        return build_tutorial_session(
            world=self.sim_world.world_state,
            world_size=float(self.world_size),
            damage_scale_by_type=self.sim_world.damage_scale_by_type,
            detail_preset=int(self._deterministic_detail_preset()),
            violence_disabled=int(self._deterministic_violence_disabled()),
            game_tune_started=bool(self.sim_world.game_tune_started),
            demo_mode_active=bool(self.demo_mode_active),
        )

    def _runtime_player_count(self) -> int:
        return 1

    def open(self) -> None:
        super().open()
        self._perk_menu.reset()

        self._skip_button = UiButtonState("Skip tutorial", force_wide=True)
        self._play_button = UiButtonState("Play a game", force_wide=True)
        self._repeat_button = UiButtonState("Repeat tutorial", force_wide=True)

        self._frame_input_state = None
        self._perk_pick_pending = False

        self.state.perk_selection.pending_count = 0
        self.state.perk_selection.choices.clear()
        self.state.perk_selection.choices_dirty = True

        status = self.state.status
        terrain = advance_unlock_terrain(
            self.state.rng,
            unlock_index=(0 if status is None else int(status.quest_unlock_index)),
            width=int(self.world_size),
            height=int(self.world_size),
        )
        self.apply_terrain_setup(
            terrain_slots=terrain.terrain_slots,
            seed=int(terrain.terrain_seed),
        )
        self.sim_world.state.rng.srand(int(self.state.rng.state))

        self.player.pos = Vec2(float(self.world_size) * 0.5, float(self.world_size) * 0.5)
        weapon_assign_player(self.player, WeaponId.PISTOL, state=self.state)
        self._sim_session = self._new_sim_session()
        self._replay_recorder = ReplayRecorder(
            ReplayHeader(
                game_mode_id=GameMode.TUTORIAL,
                seed=int(self._run_reset_seed),
                tick_rate=int(self._gameplay_tick_rate()),
                quest_fail_retry_count=int(self.quest_fail_retry_count),
                hardcore=bool(self.hardcore),
                preserve_bugs=bool(self.state.preserve_bugs),
                detail_preset=int(self._deterministic_detail_preset()),
                violence_disabled=int(self._deterministic_violence_disabled()),
                world_size=float(self.world_size),
                player_count=1,
                status=GameStatusData() if status is None else status.as_data(),
            ),
        )
        self._replay_checkpoints_sample_rate = int(DEFAULT_CHECKPOINT_SAMPLE_RATE)
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

    def close(self) -> None:
        self._sim_session = None
        self._replay_recorder = None
        self._frame_input_state = None
        super().close()

    def _replay_claimed_stats_complete(self) -> bool:
        return int(self.state.tutorial.stage_index) >= 8

    def _replay_claimed_stats_elapsed_ms(self) -> int:
        session = self._sim_session
        if session is None:
            return 0
        return int(session.elapsed_ms)

    def _replay_output_basename(self, *, stamp: str, replay) -> str:
        _ = replay
        return f"tutorial_{stamp}"

    def _open_perk_menu(self) -> None:
        self._open_perk_menu_ui(
            menu=self._perk_menu,
            players=[self.player],
            game_mode=GameMode.TUTORIAL,
            player_count=1,
        )

    def _handle_input(self) -> None:
        if self._perk_menu.open and rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._perk_menu.close()
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _build_input(self) -> PlayerInput:
        controls = self.config.controls.player(0)
        move_forward_key, move_backward_key, turn_left_key, turn_right_key = controls.move_codes
        fire_key = controls.fire_code

        move = Vec2(
            float(input_code_is_down(turn_right_key)) - float(input_code_is_down(turn_left_key)),
            float(input_code_is_down(move_backward_key)) - float(input_code_is_down(move_forward_key)),
        )

        mouse = self._ui_mouse_pos()
        aim = self.screen_to_world(Vec2.from_xy(mouse))

        fire_down = input_code_is_down(fire_key)
        fire_pressed = input_code_is_pressed(fire_key)
        reload_key = self.config.controls.reload_code
        reload_pressed = input_code_is_pressed(reload_key)

        return PlayerInput(
            move=move,
            aim=aim,
            fire_down=bool(fire_down),
            fire_pressed=bool(fire_pressed),
            reload_pressed=bool(reload_pressed),
        )

    def _build_local_inputs(self, *, dt: float) -> list[PlayerInput]:
        _ = dt
        frame_input_state = self._frame_input_state
        if frame_input_state is None:
            frame_input_state = self._build_input()
        return [frame_input_state]

    def _finish_tutorial_run(self, *, restart: bool) -> None:
        self._save_replay()
        if restart:
            self.open()
            return
        self.close_requested = True

    def _update_prompt_buttons(self, *, dt_ms: float, mouse: rl.Vector2, click: bool) -> None:
        tutorial = self.state.tutorial
        overlay = self.state.tutorial_overlay
        font = self._small
        assert font is not None, "tutorial buttons require loaded small font"
        stage = int(tutorial.stage_index)
        prompt_alpha = float(overlay.prompt_alpha)
        if stage == 8:
            self._play_button.alpha = prompt_alpha
            self._repeat_button.alpha = prompt_alpha
            self._play_button.enabled = prompt_alpha > 1e-3
            self._repeat_button.enabled = prompt_alpha > 1e-3
        else:
            skip_alpha = clamp(float(tutorial.stage_timer_ms - 1000) * 0.001, 0.0, 1.0)
            self._skip_button.alpha = skip_alpha
            self._skip_button.enabled = skip_alpha > 1e-3

        if stage == 8:
            resources = self.render_resources.resources
            rect, _lines, _line_h = tutorial_prompt_panel_rect(
                overlay.prompt_text,
                measure_text_width=lambda text, scale: float(self._ui_text_width(text, scale)),
                measure_line_height=self._ui_line_height,
                pos=TUTORIAL_PANEL_POS,
                scale=1.0,
            )
            gap = 18.0
            button_base_pos = Vec2(rect.x + 10.0, rect.y + rect.height + 10.0)
            play_w = button_width(resources, self._play_button.label, scale=1.0, force_wide=True)
            repeat_w = button_width(resources, self._repeat_button.label, scale=1.0, force_wide=True)
            if button_update(
                self._play_button,
                pos=button_base_pos,
                width=play_w,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
            ):
                self._finish_tutorial_run(restart=False)
                return
            if button_update(
                self._repeat_button,
                pos=button_base_pos.offset(dx=play_w + gap),
                width=repeat_w,
                dt_ms=dt_ms,
                mouse=mouse,
                click=click,
            ):
                self._finish_tutorial_run(restart=True)
                return
            return

        if self._skip_button.enabled:
            resources = self.render_resources.resources
            y = float(rl.get_screen_height()) - 50.0
            w = button_width(resources, self._skip_button.label, scale=1.0, force_wide=True)
            if button_update(self._skip_button, pos=Vec2(10.0, y), width=w, dt_ms=dt_ms, mouse=mouse, click=click):
                self._finish_tutorial_run(restart=False)

    def update(self, dt: float) -> None:
        self._update_audio(dt)
        dt, dt_ui_ms = self._tick_frame(dt, clamp_cursor_pulse=True)

        self._handle_input()
        if self._action == "open_pause_menu":
            return
        if self.close_requested:
            return

        perk_pending = int(self.state.perk_selection.pending_count) > 0 and self.player.health > 0.0
        choices = perk_selection_prepared_choices(self.sim_world.players, self.state.perk_selection)
        if int(self.state.tutorial.stage_index) == 6 and perk_pending and (not self._perk_menu.active) and (
            not self._perk_pick_pending
        ):
            self._open_perk_menu()
        if self._perk_menu.open:
            choice_index = self._perk_menu.handle_input(
                self._perk_menu_ui_context(),
                choices,
                dt_ui_ms=dt_ui_ms,
            )
            if choice_index is not None:
                self._perk_pick_pending = True
                self.record_perk_pick_command(int(choice_index), player_index=0)
        self._perk_menu.tick_timeline(dt_ui_ms)

        perk_menu_active = self._perk_menu.active

        dt_world = 0.0 if self._paused or perk_menu_active else dt

        input_state = self._build_input()
        if dt_world > 0.0:
            session = self._sim_session
            if session is not None:
                elapsed_before_ms = float(session.elapsed_ms)
                session.detail_preset = int(self._deterministic_detail_preset())
                session.violence_disabled = int(self._deterministic_violence_disabled())
                self._frame_input_state = input_state
                try:
                    self._run_deterministic_session_ticks(
                        dt_frame=float(dt_world),
                        session=session,
                        recorder=self._replay_recorder,
                        on_tick=lambda _tick, _tick_index: False,
                        on_checkpoint=lambda tick_index, tick: self._record_replay_checkpoint_from_tick(
                            tick_index=int(tick_index),
                            tick=tick,
                        ),
                    )
                finally:
                    self._frame_input_state = None
                if float(session.elapsed_ms) != elapsed_before_ms:
                    self._perk_pick_pending = False

        mouse = self._ui_mouse_pos()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        self._update_prompt_buttons(dt_ms=dt_ui_ms, mouse=mouse, click=click)

    def draw(self) -> None:
        perk_menu_active = self._perk_menu.active
        self._draw_world(
            draw_aim_indicators=not perk_menu_active,
            entity_alpha=self._world_entity_alpha(),
        )
        self._draw_screen_fade()

        hud_bottom = 0.0
        if not perk_menu_active:
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                HudRenderContext(
                    resources=self.render_resources.resources,
                    state=self._hud_state,
                    font=self._small,
                    alpha=1.0,
                    show_health=hud_flags.show_health,
                    show_weapon=hud_flags.show_weapon,
                    show_xp=hud_flags.show_xp,
                    show_time=hud_flags.show_time,
                    show_quest_hud=hud_flags.show_quest_hud,
                    small_indicators=self._hud_small_indicators(),
                ),
                player=self.player,
                players=self.sim_world.players,
                bonus_hud=self.state.bonus_hud,
                elapsed_ms=float(self._session_elapsed_ms() if self._sim_session is not None else 0.0),
                score=int(self.player.experience),
                frame_dt_ms=self._last_dt_ms,
            )

        self._draw_tutorial_prompts(hud_bottom=hud_bottom)

        if perk_menu_active:
            self._perk_menu.draw(
                self._perk_menu_ui_context(),
                perk_selection_prepared_choices(self.sim_world.players, self.state.perk_selection),
            )
            self._draw_menu_cursor()

    def _draw_tutorial_prompts(self, *, hud_bottom: float) -> None:
        overlay = self.state.tutorial_overlay
        draw_tutorial_overlay_panels(
            overlay,
            draw_text=lambda text, pos, color, scale: self._draw_ui_text(text, pos, color, scale=scale),
            measure_text_width=lambda text, scale: float(self._ui_text_width(text, scale)),
            measure_line_height=self._ui_line_height,
        )
        resources = self.render_resources.resources
        font = self._small
        assert font is not None, "tutorial prompts require loaded small font"

        stage = int(self.state.tutorial.stage_index)
        if stage == 8:
            rect, _lines, _line_h = tutorial_prompt_panel_rect(
                overlay.prompt_text,
                measure_text_width=lambda text, scale: float(self._ui_text_width(text, scale)),
                measure_line_height=self._ui_line_height,
                pos=TUTORIAL_PANEL_POS,
                scale=1.0,
            )
            gap = 18.0
            button_base_pos = Vec2(rect.x + 10.0, rect.y + rect.height + 10.0)
            play_w = button_width(resources, self._play_button.label, scale=1.0, force_wide=True)
            repeat_w = button_width(resources, self._repeat_button.label, scale=1.0, force_wide=True)
            button_draw(
                resources,
                self._play_button,
                pos=button_base_pos,
                width=play_w,
                scale=1.0,
            )
            button_draw(
                resources,
                self._repeat_button,
                pos=button_base_pos.offset(dx=play_w + gap),
                width=repeat_w,
                scale=1.0,
            )
            return

        if self._skip_button.alpha > 1e-3:
            y = float(rl.get_screen_height()) - 50.0
            w = button_width(resources, self._skip_button.label, scale=1.0, force_wide=True)
            button_draw(resources, self._skip_button, pos=Vec2(10.0, y), width=w, scale=1.0)

        if self._paused:
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            self._draw_ui_text("paused (TAB)", Vec2(x, y), UI_HINT_COLOR)

    def _draw_menu_cursor(self) -> None:
        resources = self.render_resources.resources
        mouse_pos = self._ui_mouse
        draw_menu_cursor(
            resources.texture(TextureId.PARTICLES),
            resources.texture(TextureId.UI_CURSOR),
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )



from __future__ import annotations

from grim.assets import RuntimeResources

from ..game.types import GameState


def require_runtime_resources(state: GameState) -> RuntimeResources:
    if state.resources is None:
        raise RuntimeError("runtime resources are not loaded")
    return state.resources



from __future__ import annotations

from grim.assets import TextureId
from grim.audio import play_sfx, update_audio
from grim.fonts.small import SmallFontData
from grim.geom import Vec2
from grim.raylib_api import rl
from grim.sfx_map import SfxId
from grim.terrain_render import GroundRenderer

from ...game.types import GameState
from ...ui.menu_panel import draw_classic_menu_panel
from ...ui.perk_menu import UiButtonState, button_draw, button_update, button_width
from ..assets import require_runtime_resources
from ..high_scores_layout import hs_left_panel_pos_x, hs_right_panel_pos_x
from ..menu import (
    MENU_PANEL_OFFSET_X,
    MENU_PANEL_OFFSET_Y,
    MENU_PANEL_WIDTH,
    MENU_SCALE_SMALL_THRESHOLD,
    MENU_SIGN_HEIGHT,
    MENU_SIGN_OFFSET_X,
    MENU_SIGN_OFFSET_Y,
    MENU_SIGN_POS_X_PAD,
    MENU_SIGN_POS_Y,
    MENU_SIGN_POS_Y_SMALL,
    MENU_SIGN_WIDTH,
    UI_SHADOW_OFFSET,
    MenuView,
    _draw_menu_cursor,
    ensure_menu_ground,
    menu_ground_camera,
)
from ..transitions import _draw_screen_fade
from .base import PANEL_TIMELINE_END_MS, PANEL_TIMELINE_START_MS

# Shared panel layout (state_14/15/16 in the oracle): tall left panel + short right panel.
LEFT_PANEL_POS_Y = 185.0
LEFT_PANEL_HEIGHT = 378.0
RIGHT_PANEL_POS_Y = 200.0
RIGHT_PANEL_HEIGHT = 254.0


class _DatabaseBaseView:
    def __init__(self, state: GameState) -> None:
        self.state = state
        self._is_open = False
        self._ground: GroundRenderer | None = None

        self._cursor_pulse_time = 0.0
        self._widescreen_y_shift = 0.0
        self._timeline_ms = 0
        self._timeline_max_ms = PANEL_TIMELINE_START_MS
        self._closing = False
        self._close_action: str | None = None
        self._pending_action: str | None = None
        self._action: str | None = None

        self._back_button = UiButtonState("Back", force_wide=False)

    def open(self) -> None:
        layout_w = float(self.state.config.display.width)
        self._widescreen_y_shift = MenuView._menu_widescreen_y_shift(layout_w)
        self._ground = None if self.state.pause_background is not None else ensure_menu_ground(self.state)
        self._cursor_pulse_time = 0.0
        self._timeline_ms = 0
        self._timeline_max_ms = PANEL_TIMELINE_START_MS
        self._closing = False
        self._close_action = None
        self._pending_action = None
        self._action = None

        self._back_button = UiButtonState("Back", force_wide=False)

        if self.state.audio is not None:
            play_sfx(self.state.audio, SfxId.UI_PANELCLICK)
        self._is_open = True

    def close(self) -> None:
        self._is_open = False
        self._ground = None
        self._closing = False
        self._close_action = None
        self._pending_action = None
        self._action = None

    def take_action(self) -> str | None:
        self._assert_open()
        if self._pending_action is not None:
            action = self._pending_action
            self._pending_action = None
            self._closing = False
            self._close_action = None
            self._timeline_ms = self._timeline_max_ms
            return action
        action = self._action
        self._action = None
        return action

    def _assert_open(self) -> None:
        assert self._is_open, f"{self.__class__.__name__} must be opened before use"

    def _panel_top_left(self, *, pos: Vec2, scale: float) -> Vec2:
        return Vec2(
            pos.x + MENU_PANEL_OFFSET_X * scale,
            pos.y + self._widescreen_y_shift + MENU_PANEL_OFFSET_Y * scale,
        )

    def _begin_close_transition(self, action: str) -> None:
        if self._closing:
            return
        self._closing = True
        self._close_action = action

    def _draw_sign(self) -> None:
        sign = require_runtime_resources(self.state).texture(TextureId.UI_SIGN_CRIMSON)
        screen_w = float(self.state.config.display.width)
        sign_scale, shift_x = MenuView._sign_layout_scale(int(screen_w))
        sign_pos = Vec2(
            screen_w + MENU_SIGN_POS_X_PAD,
            MENU_SIGN_POS_Y if screen_w > MENU_SCALE_SMALL_THRESHOLD else MENU_SIGN_POS_Y_SMALL,
        )
        sign_w = MENU_SIGN_WIDTH * sign_scale
        sign_h = MENU_SIGN_HEIGHT * sign_scale
        offset_x = MENU_SIGN_OFFSET_X * sign_scale + shift_x
        offset_y = MENU_SIGN_OFFSET_Y * sign_scale
        rotation_deg = 0.0
        fx_detail = self.state.config.display.fx_detail_enabled(level=0, default=False)
        if fx_detail:
            MenuView._draw_ui_quad_shadow(
                texture=sign,
                src=rl.Rectangle(0.0, 0.0, float(sign.width), float(sign.height)),
                dst=rl.Rectangle(sign_pos.x + UI_SHADOW_OFFSET, sign_pos.y + UI_SHADOW_OFFSET, sign_w, sign_h),
                origin=rl.Vector2(-offset_x, -offset_y),
                rotation_deg=rotation_deg,
            )
        MenuView._draw_ui_quad(
            texture=sign,
            src=rl.Rectangle(0.0, 0.0, float(sign.width), float(sign.height)),
            dst=rl.Rectangle(sign_pos.x, sign_pos.y, sign_w, sign_h),
            origin=rl.Vector2(-offset_x, -offset_y),
            rotation_deg=rotation_deg,
            tint=rl.WHITE,
        )

    def update(self, dt: float) -> None:
        self._assert_open()
        if self.state.audio is not None:
            update_audio(self.state.audio, dt)
        if self._ground is not None:
            self._ground.process_pending()
        self._cursor_pulse_time += min(float(dt), 0.1) * 1.1

        dt_ms = int(min(float(dt), 0.1) * 1000.0)
        if self._closing:
            if dt_ms > 0 and self._pending_action is None:
                self._timeline_ms -= dt_ms
                if self._timeline_ms < 0 and self._close_action is not None:
                    self._pending_action = self._close_action
                    self._close_action = None
            return

        if dt_ms > 0:
            self._timeline_ms = min(self._timeline_max_ms, int(self._timeline_ms + dt_ms))

        enabled = self._timeline_ms >= self._timeline_max_ms

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE) and enabled:
            if self.state.audio is not None:
                play_sfx(self.state.audio, SfxId.UI_BUTTONCLICK)
            self._begin_close_transition("back_to_previous")
            return

        if not enabled:
            return

        screen_width = float(self.state.config.display.width)
        scale = 1.0
        left_panel_pos_x = hs_left_panel_pos_x(screen_width)
        left_top_left = self._panel_top_left(pos=Vec2(left_panel_pos_x, LEFT_PANEL_POS_Y), scale=scale)
        resources = require_runtime_resources(self.state)

        mouse = rl.get_mouse_position()
        click = rl.is_mouse_button_pressed(rl.MouseButton.MOUSE_BUTTON_LEFT)
        self._update_content_interaction(left_top_left=left_top_left, scale=scale, mouse=mouse)

        back_pos = self._back_button_pos()
        back_w = button_width(resources, self._back_button.label, scale=scale, force_wide=self._back_button.force_wide)
        if button_update(
            self._back_button,
            pos=left_top_left + back_pos * scale,
            width=back_w,
            dt_ms=dt_ms,
            mouse=mouse,
            click=click,
        ):
            if self.state.audio is not None:
                play_sfx(self.state.audio, SfxId.UI_BUTTONCLICK)
            self._begin_close_transition("back_to_previous")

    def draw(self) -> None:
        self._assert_open()
        rl.clear_background(rl.BLACK)
        pause_background = self.state.pause_background
        if pause_background is not None:
            pause_background.draw_pause_background()
        elif self._ground is not None:
            self._ground.draw(menu_ground_camera(self.state))
        _draw_screen_fade(self.state)

        screen_width = float(self.state.config.display.width)
        scale = 1.0
        fx_detail = self.state.config.display.fx_detail_enabled(level=0, default=False)

        panel_w = MENU_PANEL_WIDTH * scale
        _angle_rad, left_slide_x = MenuView._ui_element_anim(
            self,
            index=1,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
            direction_flag=0,
        )
        _angle_rad, right_slide_x = MenuView._ui_element_anim(
            self,
            index=2,
            start_ms=PANEL_TIMELINE_START_MS,
            end_ms=PANEL_TIMELINE_END_MS,
            width=panel_w,
            direction_flag=1,
        )

        left_panel_pos_x = hs_left_panel_pos_x(screen_width)
        left_top_left = self._panel_top_left(pos=Vec2(left_panel_pos_x, LEFT_PANEL_POS_Y), scale=scale)
        right_panel_pos_x = hs_right_panel_pos_x(screen_width)
        right_top_left = self._panel_top_left(pos=Vec2(right_panel_pos_x, RIGHT_PANEL_POS_Y), scale=scale)
        left_panel_top_left = left_top_left.offset(dx=float(left_slide_x))
        right_panel_top_left = right_top_left.offset(dx=float(right_slide_x))

        draw_classic_menu_panel(
            require_runtime_resources(self.state).texture(TextureId.UI_MENU_PANEL),
            dst=rl.Rectangle(left_panel_top_left.x, left_panel_top_left.y, panel_w, LEFT_PANEL_HEIGHT * scale),
            tint=rl.WHITE,
            shadow=fx_detail,
        )
        draw_classic_menu_panel(
            require_runtime_resources(self.state).texture(TextureId.UI_MENU_PANEL),
            dst=rl.Rectangle(right_panel_top_left.x, right_panel_top_left.y, panel_w, RIGHT_PANEL_HEIGHT * scale),
            tint=rl.WHITE,
            shadow=fx_detail,
            flip_x=True,
        )

        resources = require_runtime_resources(self.state)
        font = resources.small_font
        self._draw_contents(left_panel_top_left, right_panel_top_left, scale=scale, font=font)

        back_pos = self._back_button_pos()
        back_w = button_width(resources, self._back_button.label, scale=scale, force_wide=self._back_button.force_wide)
        button_draw(
            resources,
            self._back_button,
            pos=left_panel_top_left + back_pos * scale,
            width=back_w,
            scale=scale,
        )

        self._draw_sign()
        _draw_menu_cursor(self.state, resources=resources, pulse_time=self._cursor_pulse_time)

    def _back_button_pos(self) -> Vec2:
        raise NotImplementedError

    def _draw_contents(
        self,
        left_top_left: Vec2,
        right_top_left: Vec2,
        *,
        scale: float,
        font: SmallFontData,
    ) -> None:
        raise NotImplementedError

    def _update_content_interaction(self, *, left_top_left: Vec2, scale: float, mouse: rl.Vector2) -> None:
        return



from __future__ import annotations

import math
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from functools import cache

import msgspec

from grim.raylib_api import rd, rl

from .geom import Vec2
from .rand import CrtRand

TERRAIN_TEXTURE_SIZE = 1024
TERRAIN_PATCH_SIZE = 128.0
TERRAIN_PATCH_OVERSCAN = 64.0
TERRAIN_CLEAR_COLOR = rl.Color(63, 56, 25, 255)
TERRAIN_BASE_TINT = rl.Color(178, 178, 178, 230)
TERRAIN_OVERLAY_TINT = rl.Color(178, 178, 178, 230)
TERRAIN_DETAIL_TINT = rl.Color(178, 178, 178, 153)
TERRAIN_DENSITY_BASE = 800
TERRAIN_DENSITY_OVERLAY = 0x23
TERRAIN_DENSITY_DETAIL = 0x0F
TERRAIN_DENSITY_SHIFT = 19
TERRAIN_ROTATION_MAX = 0x13A

# Grim2D enables alpha test globally with:
#   ALPHATESTENABLE=1, ALPHAFUNC=GREATER, ALPHAREF=4
# See: analysis/ghidra/raw/grim.dll_decompiled.c (FUN_10004520).
#
# raylib does not expose fixed-function alpha test, so we emulate it with a tiny
# discard shader for stamping into the terrain render target. This shim is
# required for parity; if it fails to compile, rendering should stop rather than
# silently drift away from the native cutoff behavior.
_ALPHA_TEST_VS_330 = r"""
#version 330

in vec3 vertexPosition;
in vec2 vertexTexCoord;
in vec4 vertexColor;

out vec2 fragTexCoord;
out vec4 fragColor;

uniform mat4 mvp;

void main() {
    fragTexCoord = vertexTexCoord;
    fragColor = vertexColor;
    gl_Position = mvp * vec4(vertexPosition, 1.0);
}
"""

_ALPHA_TEST_FS_330 = r"""
#version 330

in vec2 fragTexCoord;
in vec4 fragColor;

uniform sampler2D texture0;
uniform vec4 colDiffuse;

out vec4 finalColor;

void main() {
    // Emulate DX8 fixed-function alpha test after stage-0 modulation:
    // stage output = texture * diffuse, then discard when alpha <= 4/255.
    vec4 texel = texture(texture0, fragTexCoord) * fragColor * colDiffuse;
    if (texel.a <= 0.0156862745) discard;
    finalColor = texel;
}
"""


@cache
def _get_alpha_test_shader() -> rl.Shader:
    shader = rl.load_shader_from_memory(_ALPHA_TEST_VS_330, _ALPHA_TEST_FS_330)
    if shader.id <= 0:
        raise RuntimeError("terrain alpha-test shader compilation returned an invalid shader id")
    return shader


@contextmanager
def _blend_custom(src_factor: int, dst_factor: int, blend_equation: int) -> Iterator[None]:
    # NOTE: raylib/rlgl tracks custom blend factors as state; some backends only
    # apply them when switching the blend mode. Set factors both before and
    # after BeginBlendMode() to ensure the current draw uses the intended values.
    rl.rl_set_blend_factors(src_factor, dst_factor, blend_equation)
    rl.begin_blend_mode(rl.BlendMode.BLEND_CUSTOM)
    rl.rl_set_blend_factors(src_factor, dst_factor, blend_equation)
    try:
        yield
    finally:
        rl.end_blend_mode()


@contextmanager
def _color_mask(*, write_alpha: bool) -> Iterator[None]:
    rl.rl_color_mask(True, True, True, bool(write_alpha))
    try:
        yield
    finally:
        rl.rl_color_mask(True, True, True, True)


@contextmanager
def _terrain_rt_blend(
    src_factor: int,
    dst_factor: int,
    blend_equation: int,
) -> Iterator[None]:
    with _color_mask(write_alpha=False):
        with _blend_custom(src_factor, dst_factor, blend_equation):
            yield


@contextmanager
def _maybe_alpha_test() -> Iterator[None]:
    shader = _get_alpha_test_shader()
    rl.begin_shader_mode(shader)
    try:
        yield
    finally:
        rl.end_shader_mode()

class GroundDecal(msgspec.Struct):
    texture: rl.Texture
    src: rl.Rectangle
    pos: Vec2
    width: float
    height: float
    rotation_rad: float = 0.0
    tint: rl.Color = rl.WHITE


class GroundCorpseDecal(msgspec.Struct):
    bodyset_frame: int
    top_left: Vec2
    size: float
    rotation_rad: float
    tint: rl.Color = rl.WHITE


class GroundRenderer(msgspec.Struct):
    texture: rl.Texture
    overlay: rl.Texture
    overlay_detail: rl.Texture
    width: int = TERRAIN_TEXTURE_SIZE
    height: int = TERRAIN_TEXTURE_SIZE
    texture_scale: float = 1.0
    texture_failed: bool = False
    render_target: rl.RenderTexture | None = None
    _render_target_ready: bool = False
    _scheduled_seed: int | None = None

    def generation_pending(self) -> bool:
        """True while a scheduled terrain generate is still pending."""
        return self._scheduled_seed is not None

    def render_target_ready(self) -> bool:
        """True when the terrain render target exists and is ready for drawing."""
        return self.render_target is not None and self._render_target_ready

    def process_pending(self) -> None:
        seed = self._scheduled_seed
        if seed is None:
            return
        self._scheduled_seed = None
        self._generate_texture(seed=seed)

    def _ensure_render_target(self) -> None:
        scale = min(max(self.texture_scale, 0.5), 4.0)
        self.texture_scale = scale

        render_w, render_h = self._render_target_size_for(scale)
        if self._load_render_target(render_w, render_h):
            self.texture_failed = False
            return

        self.texture_failed = True
        if self.render_target is not None:
            rl.unload_render_texture(self.render_target)
            self.render_target = None
        self._render_target_ready = False

    def schedule_generate(self, seed: int) -> None:
        self._scheduled_seed = seed

    def _generate_texture(self, seed: int) -> None:
        self._ensure_render_target()
        if self.render_target is None:
            return
        rng = CrtRand(seed)
        rl.begin_texture_mode(self.render_target)
        rl.clear_background(TERRAIN_CLEAR_COLOR)
        # Intentional rewrite deviation: the classic game appears to point-sample
        # terrain stamps while rotating them into the RT, but bilinear sampling
        # reads better in the port and still stays within current fixture tolerances.
        # Keep the ground RT alpha opaque like the original exe's XRGB-style RT.
        # The port does that by masking out alpha writes while stamping.
        with _maybe_alpha_test():
            with _terrain_rt_blend(
                rd.RL_SRC_ALPHA,
                rd.RL_ONE_MINUS_SRC_ALPHA,
                rd.RL_FUNC_ADD,
            ):
                self._scatter_texture(self.texture, TERRAIN_BASE_TINT, rng, TERRAIN_DENSITY_BASE)
                self._scatter_texture(self.overlay, TERRAIN_OVERLAY_TINT, rng, TERRAIN_DENSITY_OVERLAY)
                self._scatter_texture(self.overlay_detail, TERRAIN_DETAIL_TINT, rng, TERRAIN_DENSITY_DETAIL)
        rl.end_texture_mode()
        self._render_target_ready = True

    def bake_decals(self, decals: Sequence[GroundDecal]) -> bool:
        if not decals:
            return False

        if self.render_target is None or not self._render_target_ready:
            return False

        inv_scale = 1.0 / self._normalized_texture_scale()
        rl.begin_texture_mode(self.render_target)
        with _maybe_alpha_test():
            with _terrain_rt_blend(
                rd.RL_SRC_ALPHA,
                rd.RL_ONE_MINUS_SRC_ALPHA,
                rd.RL_FUNC_ADD,
            ):
                for decal in decals:
                    w = decal.width * inv_scale
                    h = decal.height * inv_scale
                    dst = rl.Rectangle(decal.pos.x * inv_scale, decal.pos.y * inv_scale, w, h)
                    origin = rl.Vector2(w * 0.5, h * 0.5)
                    rl.draw_texture_pro(
                        decal.texture,
                        decal.src,
                        dst,
                        origin,
                        math.degrees(decal.rotation_rad),
                        decal.tint,
                    )
        rl.end_texture_mode()

        self._render_target_ready = True
        return True

    def bake_corpse_decals(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
    ) -> bool:
        if not decals:
            return False

        if self.render_target is None or not self._render_target_ready:
            return False

        scale = self._normalized_texture_scale()
        inv_scale = 1.0 / scale
        offset = 2.0 * scale / float(self.width)
        rl.begin_texture_mode(self.render_target)
        # Intentional rewrite deviation: the classic game appears to point-sample
        # corpse atlas frames while baking, but bilinear sampling reads better in
        # the port at modern output scales.
        with _maybe_alpha_test():
            self._draw_corpse_shadow_pass(bodyset_texture, decals, inv_scale, offset)
            self._draw_corpse_color_pass(bodyset_texture, decals, inv_scale, offset)
        rl.end_texture_mode()

        self._render_target_ready = True
        return True

    def draw(self, camera: Vec2) -> None:
        out_w = max(1.0, float(rl.get_screen_width()))
        out_h = max(1.0, float(rl.get_screen_height()))
        screen_w, screen_h = self._fit_view_window(out_w, out_h)
        cam = self._clamp_camera(camera, screen_w, screen_h)
        self._draw_view(cam, screen_w=screen_w, screen_h=screen_h, out_w=out_w, out_h=out_h)

    def draw_view(
        self,
        camera: Vec2,
        *,
        screen_w: float,
        screen_h: float,
        out_w: float,
        out_h: float,
    ) -> None:
        self._draw_view(
            camera,
            screen_w=max(1.0, float(screen_w)),
            screen_h=max(1.0, float(screen_h)),
            out_w=max(1.0, float(out_w)),
            out_h=max(1.0, float(out_h)),
        )

    def _draw_view(
        self,
        camera: Vec2,
        *,
        screen_w: float,
        screen_h: float,
        out_w: float,
        out_h: float,
    ) -> None:
        if self.render_target is None or not self._render_target_ready:
            rl.draw_rectangle(0, 0, int(out_w + 0.5), int(out_h + 0.5), TERRAIN_CLEAR_COLOR)
            return

        target = self.render_target
        u0 = -camera.x / float(self.width)
        v0 = -camera.y / float(self.height)
        u1 = u0 + screen_w / float(self.width)
        v1 = v0 + screen_h / float(self.height)
        src_x = u0 * float(target.texture.width)
        # Render textures are vertically flipped in raylib, so adjust the source
        # rectangle to sample the correct world-space slice before flipping.
        src_y = (1.0 - v1) * float(target.texture.height)
        src_w = (u1 - u0) * float(target.texture.width)
        src_h = (v1 - v0) * float(target.texture.height)
        src = rl.Rectangle(src_x, src_y, src_w, -src_h)
        dst = rl.Rectangle(0.0, 0.0, out_w, out_h)
        # Disable alpha blending when drawing terrain to screen - the render target's
        # alpha channel may be < 1.0 after stamp blending, but terrain should be opaque.
        with _blend_custom(rd.RL_ONE, rd.RL_ZERO, rd.RL_FUNC_ADD):
            rl.draw_texture_pro(target.texture, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)

    def _fit_view_window(self, screen_w: float, screen_h: float) -> tuple[float, float]:
        """
        Convert output dimensions into a world-space camera window.

        Keep a uniform pixel scale and never request a camera window larger than
        the terrain dimensions. This avoids non-uniform stretch on widescreen
        outputs where only one axis exceeds world size.
        """

        world_w = float(self.width)
        world_h = float(self.height)
        if world_w <= 0.0 or world_h <= 0.0:
            return max(1.0, float(screen_w)), max(1.0, float(screen_h))

        out_w = max(1.0, float(screen_w))
        out_h = max(1.0, float(screen_h))
        scale = max(out_w / world_w, out_h / world_h, 1.0)
        view_w = min(world_w, out_w / scale)
        view_h = min(world_h, out_h / scale)
        return view_w, view_h

    def _scatter_texture(
        self,
        texture: rl.Texture,
        tint: rl.Color,
        rng: CrtRand,
        density: int,
    ) -> None:
        area = self.width * self.height
        count = (area * density) >> TERRAIN_DENSITY_SHIFT
        if count <= 0:
            return
        inv_scale = 1.0 / self._normalized_texture_scale()
        size = TERRAIN_PATCH_SIZE * inv_scale
        src = rl.Rectangle(0.0, 0.0, float(texture.width), float(texture.height))
        origin = rl.Vector2(size * 0.5, size * 0.5)
        span_w = self.width + int(TERRAIN_PATCH_OVERSCAN * 2)
        # The original exe uses `terrain_texture_width` for both axes. Terrain is
        # square (1024x1024) so this is equivalent, but keep it for parity.
        span_h = span_w
        for _ in range(count):
            angle = ((rng.rand() % TERRAIN_ROTATION_MAX) * 0.01) % math.tau
            # IMPORTANT: The exe consumes RNG as rotation, then Y, then X.
            y = ((rng.rand() % span_h) - TERRAIN_PATCH_OVERSCAN) * inv_scale
            x = ((rng.rand() % span_w) - TERRAIN_PATCH_OVERSCAN) * inv_scale
            # raylib's DrawTexturePro positions the quad by the *origin point*,
            # while the original engine uses x/y as the quad top-left.
            dst = rl.Rectangle(float(x + size * 0.5), float(y + size * 0.5), size, size)
            rl.draw_texture_pro(texture, src, dst, origin, math.degrees(angle), tint)

    def _clamp_camera(self, camera: Vec2, screen_w: float, screen_h: float) -> Vec2:
        min_x = screen_w - float(self.width)
        min_y = screen_h - float(self.height)
        return Vec2(
            max(min(camera.x, -1.0), min_x),
            max(min(camera.y, -1.0), min_y),
        )

    def _load_render_target(self, render_w: int, render_h: int) -> bool:
        if self.render_target is not None:
            if self.render_target.texture.width == render_w and self.render_target.texture.height == render_h:
                return True
            rl.unload_render_texture(self.render_target)
            self.render_target = None
            self._render_target_ready = False

        candidate = rl.load_render_texture(render_w, render_h)
        if candidate.id <= 0 or candidate.texture.width <= 0 or candidate.texture.height <= 0:
            rl.unload_render_texture(candidate)
            return False
        # raylib's IsRenderTextureValid() only checks ids and dimensions; use the
        # rlgl completeness check so incomplete FBO attachments fail immediately.
        if not rl.rl_framebuffer_complete(candidate.id):
            rl.unload_render_texture(candidate)
            return False

        self.render_target = candidate
        self._render_target_ready = False
        rl.set_texture_filter(self.render_target.texture, rl.TextureFilter.TEXTURE_FILTER_BILINEAR)
        rl.set_texture_wrap(self.render_target.texture, rl.TextureWrap.TEXTURE_WRAP_CLAMP)
        return True

    def _render_pixel_ratio(self) -> float:
        screen_w = int(rl.get_screen_width())
        screen_h = int(rl.get_screen_height())
        render_w = int(rl.get_render_width())
        render_h = int(rl.get_render_height())
        if render_w == screen_w * 2 and render_h == screen_h * 2:
            return 2.0
        return 1.0

    def _render_target_size_for(self, scale: float) -> tuple[int, int]:
        pixel_scale = self._render_pixel_ratio()
        render_w = max(1, int((self.width * pixel_scale) / scale))
        render_h = max(1, int((self.height * pixel_scale) / scale))
        return render_w, render_h

    def _normalized_texture_scale(self) -> float:
        scale = self.texture_scale
        if scale < 0.5:
            scale = 0.5
        if self._render_pixel_ratio() == 2.0:
            scale *= 0.5
        return scale

    def _corpse_src(self, bodyset_texture: rl.Texture, frame: int) -> rl.Rectangle:
        frame = int(frame) & 0xF
        cell_w = float(bodyset_texture.width) * 0.25
        cell_h = float(bodyset_texture.height) * 0.25
        col = frame & 3
        row = frame >> 2
        return rl.Rectangle(cell_w * float(col), cell_h * float(row), cell_w, cell_h)

    def _draw_corpse_shadow_pass(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
        inv_scale: float,
        offset: float,
    ) -> None:
        with _terrain_rt_blend(
            rd.RL_ZERO,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_FUNC_ADD,
        ):
            for decal in decals:
                src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                size = decal.size * inv_scale * 1.064
                x = (decal.top_left.x - 0.5) * inv_scale - offset
                y = (decal.top_left.y - 0.5) * inv_scale - offset
                dst = rl.Rectangle(x + size * 0.5, y + size * 0.5, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                tint = rl.Color(
                    decal.tint.r,
                    decal.tint.g,
                    decal.tint.b,
                    int(decal.tint.a * 0.5),
                )
                rl.draw_texture_pro(
                    bodyset_texture,
                    src,
                    dst,
                    origin,
                    math.degrees(decal.rotation_rad - (math.pi * 0.5)),
                    tint,
                )

    def _draw_corpse_color_pass(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
        inv_scale: float,
        offset: float,
    ) -> None:
        with _terrain_rt_blend(
            rd.RL_SRC_ALPHA,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_FUNC_ADD,
        ):
            for decal in decals:
                src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                size = decal.size * inv_scale
                x = decal.top_left.x * inv_scale - offset
                y = decal.top_left.y * inv_scale - offset
                dst = rl.Rectangle(x + size * 0.5, y + size * 0.5, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                rl.draw_texture_pro(
                    bodyset_texture,
                    src,
                    dst,
                    origin,
                    math.degrees(decal.rotation_rad - (math.pi * 0.5)),
                    decal.tint,
                )



from __future__ import annotations

import math
from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING

import msgspec

from grim.geom import Vec2

if TYPE_CHECKING:
    from ...creatures.runtime import CreatureState

_SPATIAL_BUCKET_SIZE = 64.0
_NATIVE_FIND_SIZE_SCALE = 0.14285715
_NATIVE_FIND_BASE_MARGIN = 3.0
_NATIVE_FIND_RADIUS_MARGIN_EPS = 0.001


def native_find_margin_for_size(size: float) -> float:
    return float(size) * _NATIVE_FIND_SIZE_SCALE + _NATIVE_FIND_BASE_MARGIN


class CreatureSpatialHash(msgspec.Struct):
    creatures: Sequence[CreatureState]
    is_collidable: Callable[[CreatureState], bool]
    bucket_size: float = _SPATIAL_BUCKET_SIZE
    _bucket_size: float = _SPATIAL_BUCKET_SIZE
    _cells: dict[tuple[int, int], list[int]] = msgspec.field(default_factory=dict)
    _cell_by_index: list[tuple[int, int] | None] = msgspec.field(default_factory=list)
    _max_find_margin: float = 0.0

    def __post_init__(self) -> None:
        if float(self.bucket_size) > 0.0:
            self._bucket_size = float(self.bucket_size)
        self.rebuild()

    def rebuild(self) -> None:
        cells: dict[tuple[int, int], list[int]] = defaultdict(list)
        cell_by_index: list[tuple[int, int] | None] = [None] * len(self.creatures)
        max_find_margin = 0.0

        for idx, creature in enumerate(self.creatures):
            if not self.is_collidable(creature):
                continue
            cell = self._cell_for_pos(creature.pos)
            cells[cell].append(int(idx))
            cell_by_index[int(idx)] = cell
            creature_find_margin = native_find_margin_for_size(float(creature.size))
            if creature_find_margin > max_find_margin:
                max_find_margin = float(creature_find_margin)

        self._cells = cells
        self._cell_by_index = cell_by_index
        self._max_find_margin = float(max_find_margin)

    def sync_index(self, index: int) -> None:
        if not (0 <= int(index) < len(self.creatures)):
            return
        idx = int(index)
        creature = self.creatures[idx]
        previous_cell = self._cell_by_index[idx]
        if not self.is_collidable(creature):
            if previous_cell is not None:
                self._remove_from_cell(idx, previous_cell)
                self._cell_by_index[idx] = None
            return

        next_cell = self._cell_for_pos(creature.pos)
        if previous_cell == next_cell:
            return
        if previous_cell is not None:
            self._remove_from_cell(idx, previous_cell)
        self._cells.setdefault(next_cell, []).append(idx)
        self._cell_by_index[idx] = next_cell

        creature_find_margin = native_find_margin_for_size(float(creature.size))
        if creature_find_margin > self._max_find_margin:
            self._max_find_margin = float(creature_find_margin)

    def candidate_indices(self, *, pos: Vec2, radius: float) -> list[int]:
        if not self._cells:
            return []
        proj_cell_x = int(math.floor(float(pos.x) / self._bucket_size))
        proj_cell_y = int(math.floor(float(pos.y) / self._bucket_size))
        max_axis_delta = float(radius) + float(self._max_find_margin) + _NATIVE_FIND_RADIUS_MARGIN_EPS
        cell_span = int(math.ceil(float(max_axis_delta) / self._bucket_size))

        candidates: list[int] = []
        for cell_y in range(proj_cell_y - cell_span, proj_cell_y + cell_span + 1):
            for cell_x in range(proj_cell_x - cell_span, proj_cell_x + cell_span + 1):
                bucket = self._cells.get((int(cell_x), int(cell_y)))
                if bucket is not None:
                    candidates.extend(bucket)

        if len(candidates) > 1:
            candidates.sort()
        return candidates

    def _cell_for_pos(self, pos: Vec2) -> tuple[int, int]:
        cell_x = int(math.floor(float(pos.x) / self._bucket_size))
        cell_y = int(math.floor(float(pos.y) / self._bucket_size))
        return (cell_x, cell_y)

    def _remove_from_cell(self, index: int, cell: tuple[int, int]) -> None:
        bucket = self._cells.get(cell)
        if bucket is None:
            return
        try:
            bucket.remove(int(index))
        except ValueError:
            return
        if not bucket:
            self._cells.pop(cell, None)


__all__ = ["CreatureSpatialHash", "native_find_margin_for_size"]



from __future__ import annotations

import msgspec

from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rd, rl

SHADER_STAMP_ANALYTIC_RADIUS_SCALE = 16.0
SHADER_STAMP_VIRTUAL_PROFILE_A = 1.3
SHADER_STAMP_VIRTUAL_PROFILE_LINEAR = -4.8
SHADER_STAMP_VIRTUAL_PROFILE_QUAD = 1.0
SHADER_STAMP_VIRTUAL_PROFILE_OFFSET = 0.01
SHADER_STAMP_VIRTUAL_INTENSITY_GAIN = 0.92
SHADER_STAMP_VIRTUAL_MAX_STAMPS = 128
SHADER_STAMP_VIRTUAL_HEAD_RADIUS_MULTIPLIER = 1.05
SHADER_STAMP_VIRTUAL_HEAD_FIRE_RADIUS_MULTIPLIER = 1.35

_BEAM_SHADER_VS_330 = """
#version 330

in vec3 vertexPosition;
in vec2 vertexTexCoord;
in vec4 vertexColor;
out vec2 fragTexCoord;
out vec4 fragColor;
out float fragLen;

uniform mat4 mvp;

void main() {
    fragTexCoord = vertexTexCoord;
    fragColor = vertexColor;
    fragLen = vertexPosition.z;
    gl_Position = mvp * vec4(vertexPosition.x, vertexPosition.y, 0.0, 1.0);
}
"""

_BEAM_FAST_STAMPED_FS_330 = f"""
#version 330

in vec2 fragTexCoord;
in vec4 fragColor;
in float fragLen;

uniform vec4 colDiffuse;
uniform float u_step_uv;
uniform float u_stamp_scale;
uniform float u_stamp_decay;
uniform float u_stamp_quad;
uniform float u_stamp_offset;
uniform float u_intensity_gain;

out vec4 finalColor;

void main() {{
    float u_len = fragLen;
    float gain = max(u_intensity_gain, 0.0);

    if (u_len < 0.0) {{
        float d = length(fragTexCoord);
        float profile = clamp(u_stamp_scale * exp(-u_stamp_decay * d - u_stamp_quad * d * d) - u_stamp_offset, 0.0, 1.0);
        float intensity = profile * fragColor.a * gain;
        vec3 rgb = fragColor.rgb * colDiffuse.rgb * intensity;
        finalColor = vec4(rgb, 1.0);
        return;
    }}

    float step_uv = max(0.001, u_step_uv);
    float len_uv = max(0.0, u_len);
    float stamp_count = floor(len_uv / step_uv) + 1.0;
    stamp_count = clamp(stamp_count, 0.0, float({SHADER_STAMP_VIRTUAL_MAX_STAMPS:d}));

    int start_i = 0;
    int end_i = int(stamp_count);

    // Profile support radius in UV-space where profile falls to zero.
    // For current params (offset > 0), this makes the loop O(local overlap).
    float reach_uv = 0.0;
    bool has_reach_uv = false;
    if (u_stamp_scale <= 0.0 || u_stamp_offset >= u_stamp_scale) {{
        has_reach_uv = true;
    }} else if (u_stamp_offset > 0.0) {{
        float target = -log(u_stamp_offset / max(u_stamp_scale, 1e-6));
        float b = max(u_stamp_decay, 0.0);
        float q = max(u_stamp_quad, 0.0);
        if (q > 1e-6) {{
            float disc = max(0.0, b * b + 4.0 * q * target);
            reach_uv = (-b + sqrt(disc)) / (2.0 * q);
            has_reach_uv = true;
        }} else if (b > 1e-6) {{
            reach_uv = target / b;
            has_reach_uv = true;
        }}
    }}

    if (has_reach_uv) {{
        reach_uv = max(reach_uv, 0.0);
        float center_index = floor(fragTexCoord.x / step_uv);
        float index_reach = ceil(reach_uv / step_uv);
        start_i = max(0, int(center_index - index_reach));
        end_i = min(int(stamp_count), int(center_index + index_reach + 1.0));
    }}

    float accum = 0.0;
    for (int i = start_i; i < end_i; i++) {{
        float fi = float(i);
        float sx = fi * step_uv;
        if (sx >= len_uv) {{
            break;
        }}
        float t = clamp(sx / max(1e-6, len_uv), 0.0, 1.0);
        float d = length(vec2(fragTexCoord.x - sx, fragTexCoord.y));
        float profile = clamp(u_stamp_scale * exp(-u_stamp_decay * d - u_stamp_quad * d * d) - u_stamp_offset, 0.0, 1.0);
        accum += t * profile;
    }}

    float intensity = accum * fragColor.a * gain;
    vec3 rgb = fragColor.rgb * colDiffuse.rgb * intensity;
    finalColor = vec4(rgb, 1.0);
}}
"""


class _BeamFastStampedShader(msgspec.Struct, frozen=True):
    shader: rl.Shader
    color_loc: int
    step_uv_loc: int
    stamp_scale_loc: int
    stamp_decay_loc: int
    stamp_quad_loc: int
    stamp_offset_loc: int
    intensity_gain_loc: int


_BEAM_FAST_STAMPED_SHADER_TRIED = False
_BEAM_FAST_STAMPED_SHADER: _BeamFastStampedShader | None = None


def _set_shader_float(shader: rl.Shader, location: int, value: float) -> None:
    if int(location) < 0:
        return
    rl.set_shader_value(
        shader,
        int(location),
        rl.ffi.new("float *", float(value)),
        rl.ShaderUniformDataType.SHADER_UNIFORM_FLOAT,
    )


def _set_shader_vec4(shader: rl.Shader, location: int, x: float, y: float, z: float, w: float) -> None:
    if int(location) < 0:
        return
    values = rl.ffi.new("float[4]", [float(x), float(y), float(z), float(w)])
    rl.set_shader_value(
        shader,
        int(location),
        values,
        rl.ShaderUniformDataType.SHADER_UNIFORM_VEC4,
    )


def _get_beam_fast_stamped_shader() -> _BeamFastStampedShader | None:
    global _BEAM_FAST_STAMPED_SHADER_TRIED, _BEAM_FAST_STAMPED_SHADER
    if _BEAM_FAST_STAMPED_SHADER_TRIED:
        return _BEAM_FAST_STAMPED_SHADER

    _BEAM_FAST_STAMPED_SHADER_TRIED = True
    try:
        shader = rl.load_shader_from_memory(_BEAM_SHADER_VS_330, _BEAM_FAST_STAMPED_FS_330)
    except (RuntimeError, OSError, ValueError):
        _BEAM_FAST_STAMPED_SHADER = None
        return None

    if int(shader.id) <= 0:
        _BEAM_FAST_STAMPED_SHADER = None
        return None

    _BEAM_FAST_STAMPED_SHADER = _BeamFastStampedShader(
        shader=shader,
        color_loc=int(rl.get_shader_location(shader, "colDiffuse")),
        step_uv_loc=int(rl.get_shader_location(shader, "u_step_uv")),
        stamp_scale_loc=int(rl.get_shader_location(shader, "u_stamp_scale")),
        stamp_decay_loc=int(rl.get_shader_location(shader, "u_stamp_decay")),
        stamp_quad_loc=int(rl.get_shader_location(shader, "u_stamp_quad")),
        stamp_offset_loc=int(rl.get_shader_location(shader, "u_stamp_offset")),
        intensity_gain_loc=int(rl.get_shader_location(shader, "u_intensity_gain")),
    )
    return _BEAM_FAST_STAMPED_SHADER


def _require_beam_fast_stamped_shader() -> _BeamFastStampedShader:
    shader_data = _get_beam_fast_stamped_shader()
    if shader_data is None:
        raise RuntimeError("rtx mode requires beam virtual shader, but it failed to load/compile")
    return shader_data


def _apply_virtual_beam_uniforms(*, shader_data: _BeamFastStampedShader, step_uv: float, intensity_gain: float) -> None:
    shader = shader_data.shader
    _set_shader_vec4(shader, shader_data.color_loc, 1.0, 1.0, 1.0, 1.0)
    _set_shader_float(shader, shader_data.step_uv_loc, float(step_uv))
    _set_shader_float(shader, shader_data.stamp_scale_loc, float(SHADER_STAMP_VIRTUAL_PROFILE_A))
    _set_shader_float(shader, shader_data.stamp_decay_loc, float(-SHADER_STAMP_VIRTUAL_PROFILE_LINEAR))
    _set_shader_float(shader, shader_data.stamp_quad_loc, float(SHADER_STAMP_VIRTUAL_PROFILE_QUAD))
    _set_shader_float(shader, shader_data.stamp_offset_loc, float(SHADER_STAMP_VIRTUAL_PROFILE_OFFSET))
    _set_shader_float(shader, shader_data.intensity_gain_loc, float(intensity_gain))


def draw_beam_fast_stamped_body(
    *,
    origin_screen: Vec2,
    head_screen: Vec2,
    start_dist_units: float,
    span_dist_units: float,
    step_units: float,
    effect_scale: float,
    scale: float,
    base_alpha: float,
    streak_rgb: tuple[float, float, float],
) -> bool:
    shader_data = _require_beam_fast_stamped_shader()

    span_units = max(0.0, float(span_dist_units))
    if span_units <= 1e-6:
        return True

    ray = head_screen - origin_screen
    direction_screen, ray_len = ray.normalized_with_length()
    if ray_len <= 1e-6:
        return True

    total_units = max(1e-6, float(start_dist_units) + span_units)
    start_t = clamp(float(start_dist_units) / total_units, 0.0, 1.0)
    p0 = origin_screen + ray * start_t
    p1 = head_screen
    direction, length = (p1 - p0).normalized_with_length()
    if length <= 1e-6:
        return True

    radius = max(0.001, float(SHADER_STAMP_ANALYTIC_RADIUS_SCALE) * float(effect_scale) * float(scale))
    screen_per_unit = float(length) / max(1e-6, span_units)
    step_screen = max(1e-6, float(step_units) * float(screen_per_unit))
    step_uv = max(1e-4, float(step_screen) / float(radius))
    u_len = float(length) / float(radius)

    side = direction.perp_left() * radius
    tail_end = p0 - direction * radius
    head_end = p1 + direction * radius

    alpha_u8 = int(clamp(float(base_alpha) * 255.0, 0.0, 255.0) + 0.5)
    if alpha_u8 <= 0:
        return True

    r = int(clamp(float(streak_rgb[0]) * 255.0, 0.0, 255.0) + 0.5)
    g = int(clamp(float(streak_rgb[1]) * 255.0, 0.0, 255.0) + 0.5)
    b = int(clamp(float(streak_rgb[2]) * 255.0, 0.0, 255.0) + 0.5)

    shader = shader_data.shader
    rl.begin_shader_mode(shader)
    _apply_virtual_beam_uniforms(
        shader_data=shader_data,
        step_uv=float(step_uv),
        intensity_gain=float(SHADER_STAMP_VIRTUAL_INTENSITY_GAIN),
    )
    rl.rl_set_texture(0)
    rl.rl_begin(rd.RL_QUADS)

    pos = tail_end - side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(-1.0, -1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), float(u_len))

    pos = tail_end + side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(-1.0, 1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), float(u_len))

    pos = head_end + side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(float(u_len + 1.0), 1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), float(u_len))

    pos = head_end - side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(float(u_len + 1.0), -1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), float(u_len))

    rl.rl_end()
    rl.rl_set_texture(0)
    rl.end_shader_mode()
    return True


def draw_beam_fast_stamped_head(
    *,
    center_screen: Vec2,
    rotation_rad: float,
    effect_scale: float,
    scale: float,
    base_alpha: float,
    head_rgb: tuple[float, float, float],
    is_fire: bool,
) -> bool:
    shader_data = _require_beam_fast_stamped_shader()

    alpha_u8 = int(clamp(float(base_alpha) * 255.0, 0.0, 255.0) + 0.5)
    if alpha_u8 <= 0:
        return True

    radius_multiplier = (
        float(SHADER_STAMP_VIRTUAL_HEAD_FIRE_RADIUS_MULTIPLIER)
        if is_fire
        else float(SHADER_STAMP_VIRTUAL_HEAD_RADIUS_MULTIPLIER)
    )
    radius = max(
        0.001,
        float(SHADER_STAMP_ANALYTIC_RADIUS_SCALE) * float(effect_scale) * float(scale) * float(radius_multiplier),
    )
    direction = Vec2.from_angle(float(rotation_rad))
    side = direction.perp_left() * radius
    forward = direction * radius

    r = int(clamp(float(head_rgb[0]) * 255.0, 0.0, 255.0) + 0.5)
    g = int(clamp(float(head_rgb[1]) * 255.0, 0.0, 255.0) + 0.5)
    b = int(clamp(float(head_rgb[2]) * 255.0, 0.0, 255.0) + 0.5)

    shader = shader_data.shader
    rl.begin_shader_mode(shader)
    _apply_virtual_beam_uniforms(
        shader_data=shader_data,
        step_uv=1.0,
        intensity_gain=float(SHADER_STAMP_VIRTUAL_INTENSITY_GAIN),
    )
    rl.rl_set_texture(0)
    rl.rl_begin(rd.RL_QUADS)

    pos = center_screen - forward - side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(-1.0, -1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), -1.0)

    pos = center_screen - forward + side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(-1.0, 1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), -1.0)

    pos = center_screen + forward + side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(1.0, 1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), -1.0)

    pos = center_screen + forward - side
    rl.rl_color4ub(int(r), int(g), int(b), int(alpha_u8))
    rl.rl_tex_coord2f(1.0, -1.0)
    rl.rl_vertex3f(float(pos.x), float(pos.y), -1.0)

    rl.rl_end()
    rl.rl_set_texture(0)
    rl.end_shader_mode()
    return True


__all__ = ["draw_beam_fast_stamped_body", "draw_beam_fast_stamped_head"]



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ..game_modes import GameMode
from ..sim.batch_apply import apply_presentation_outputs, apply_sim_metadata_batch
from ..sim.clock import FixedStepClock
from ..sim.frame_pump import advance_tick_runner_frame
from ..sim.input import PlayerInput
from ..sim.input_providers import FrameContext, LocalInputProvider
from ..sim.presentation_reactions import (
    PostApplyReaction,
    apply_post_apply_reaction,
    build_post_apply_reaction,
)
from ..sim.sessions import DeterministicSession
from ..sim.tick_runner import TickBatchResult, TickRunner, TickRunnerConfig

if TYPE_CHECKING:
    from .runtime import WorldRuntime


WorldTickInputBuilder = Callable[[FrameContext], Sequence[PlayerInput]]


@dataclass(slots=True)
class StandaloneTickHarness:
    """Standalone local runner for demo/debug screens outside BaseGameplayMode."""

    game_mode: GameMode
    build_inputs: WorldTickInputBuilder
    tick_rate: int = 60
    session: DeterministicSession | None = None
    runner: TickRunner | None = None
    world_state: object | None = None
    player_count: int = 0
    clock: FixedStepClock = field(init=False)
    frame_index: int = 0
    next_tick_index: int = 0

    def __post_init__(self) -> None:
        self.tick_rate = max(1, int(self.tick_rate))
        self.clock = FixedStepClock(tick_rate=int(self.tick_rate))

    def reset(self) -> None:
        self.session = None
        self.runner = None
        self.world_state = None
        self.player_count = 0
        self.clock = FixedStepClock(tick_rate=int(self.tick_rate))
        self.frame_index = 0
        self.next_tick_index = 0

    def _ensure_runner(self, runtime: WorldRuntime) -> tuple[TickRunner, DeterministicSession]:
        world_state = runtime.sim_world.world_state
        player_count = len(runtime.sim_world.players)
        session = self.session
        runner = self.runner
        if (
            session is not None
            and runner is not None
            and self.world_state is world_state
            and int(self.player_count) == int(player_count)
        ):
            return runner, session

        detail_preset = 5
        violence_disabled = 0
        config = runtime.config
        if config is not None:
            detail_preset = config.display.detail_preset
            violence_disabled = config.display.violence_disabled

        session = DeterministicSession(
            world=world_state,
            world_size=float(runtime.world_size),
            damage_scale_by_type=runtime.sim_world.damage_scale_by_type,
            game_mode=self.game_mode,
            detail_preset=int(detail_preset),
            violence_disabled=int(violence_disabled),
            game_tune_started=bool(runtime.sim_world.game_tune_started),
            demo_mode_active=bool(runtime.demo_mode_active),
            perk_progression_enabled=False,
            apply_world_dt_steps=True,
        )
        provider = LocalInputProvider(
            player_count=int(player_count),
            build_inputs=self.build_inputs,
        )
        runner = TickRunner(
            session=session,
            input_provider=provider,
            config=TickRunnerConfig(),
        )
        self.session = session
        self.runner = runner
        self.world_state = world_state
        self.player_count = int(player_count)
        self.clock = FixedStepClock(tick_rate=int(self.tick_rate))
        self.frame_index = 0
        self.next_tick_index = 0
        return runner, session

    def _apply_tick_batch(
        self,
        runtime: WorldRuntime,
        *,
        batch: TickBatchResult,
        session: DeterministicSession,
    ) -> int:
        outputs = apply_sim_metadata_batch(
            sim_world=runtime.sim_world,
            completed_results=batch.completed_results,
            game_tune_started=bool(session.game_tune_started),
        )
        reactions = {
            int(result.source_tick.tick_index): build_post_apply_reaction(tick_result=result)
            for result in batch.completed_results
        }
        apply_presentation_outputs(
            outputs=outputs,
            sync_audio_bridge_state=runtime.sync_audio_bridge_state,
            apply_audio_plan=lambda plan, should_apply_audio: runtime.audio_bridge.apply_plan(
                plan=plan,
                apply_audio=bool(should_apply_audio),
            ),
            apply_terrain_fx=runtime.render_resources.consume_terrain_fx_batch,
            update_camera=runtime.update_camera,
            on_output_applied=lambda output: apply_post_apply_reaction(
                reaction=reactions.get(int(output.tick_index), PostApplyReaction()),
                play_sfx=runtime.audio_bridge.router.play_sfx,
            ),
            apply_audio=True,
        )
        return len(outputs)

    def advance_frame(self, runtime: WorldRuntime, dt: float) -> int:
        if not runtime.sim_world.players:
            return 0
        runtime.terrain_runtime.process_pending()
        runner, session = self._ensure_runner(runtime)
        session.demo_mode_active = bool(runtime.demo_mode_active)
        dt = float(dt)
        ticks_requested = int(self.clock.advance(dt))
        advance = advance_tick_runner_frame(
            runner=runner,
            start_tick=int(self.next_tick_index),
            frame_index=int(self.frame_index),
            ticks_requested=int(ticks_requested),
            dt_seconds=float(dt),
            tick_dt_seconds=float(self.clock.dt_tick),
            is_networked=False,
            is_replay=False,
            refund_clock=self.clock,
        )
        self.frame_index = int(advance.frame_index)
        self.next_tick_index = int(advance.next_tick_index)
        return self._apply_tick_batch(
            runtime,
            batch=advance.batch,
            session=session,
        )



from __future__ import annotations

from pathlib import Path

from grim.audio import AudioState
from grim.config import CrimsonConfig
from grim.geom import Vec2
from grim.rand import Crand
from grim.raylib_api import rl

from ..render.frame import RenderFrame
from ..render.rtx.mode import RtxRenderMode
from ..render.world import viewport
from ..render.world.renderer import WorldRenderer
from .audio_bridge import AudioBridge
from .render_resources import RenderResources
from .sim_world_state import SimWorldState
from .terrain_runtime import TerrainRuntime


class WorldRuntime:
    """Composition container owning the 4 world components and shared lifecycle methods.
    """

    def __init__(
        self,
        *,
        assets_dir: Path,
        world_size: float = 1024.0,
        demo_mode_active: bool = False,
        quest_fail_retry_count: int = 0,
        hardcore: bool = False,
        preserve_bugs: bool = False,
        config: CrimsonConfig | None = None,
        audio_rng: Crand,
        audio: AudioState | None = None,
        rtx_mode: RtxRenderMode = RtxRenderMode.CLASSIC,
    ) -> None:
        self.assets_dir = Path(assets_dir)
        self.world_size = float(world_size)
        self.demo_mode_active = bool(demo_mode_active)
        self.quest_fail_retry_count = int(quest_fail_retry_count)
        self.hardcore = bool(hardcore)
        self.preserve_bugs = bool(preserve_bugs)
        self.config = config
        self.audio = audio
        self.audio_rng = audio_rng
        self.rtx_mode = rtx_mode

        self.sim_world = SimWorldState(
            world_size=float(self.world_size),
            demo_mode_active=bool(self.demo_mode_active),
            hardcore=bool(self.hardcore),
            quest_fail_retry_count=int(self.quest_fail_retry_count),
            preserve_bugs=bool(self.preserve_bugs),
        )
        render_resources = RenderResources(
            assets_dir=self.assets_dir,
            world_size=float(self.world_size),
            config=self.config,
        )
        self.render_resources = render_resources
        self.audio_bridge = AudioBridge(
            demo_mode_active=bool(self.demo_mode_active),
            reflex_boost_timer_source=lambda: float(self.sim_world.state.bonuses.reflex_boost),
            audio=self.audio,
            audio_rng=self.audio_rng,
        )
        self.terrain_runtime = TerrainRuntime(
            world_size=float(self.world_size),
            render_resources=render_resources,
        )

        self.camera = Vec2(-1.0, -1.0)
        self.lan_player_rings_enabled = False
        self.lan_local_aim_indicators_only = False
        self.lan_local_player_slot_index = 0
        self.renderer = WorldRenderer(
            world_size=float(self.world_size),
            config=self.config,
            camera=self.camera,
        )

        self._sync_world_size_ownership()
        self.sync_audio_bridge_state()

    # ------------------------------------------------------------------
    # Shared lifecycle methods (extracted from 4 identical implementations)
    # ------------------------------------------------------------------

    def sync_world_size(self) -> None:
        self._sync_world_size_ownership()

    def _sync_world_size_ownership(self) -> None:
        world_size = float(self.world_size)
        self.sim_world.world_size = world_size
        self.render_resources.world_size = world_size
        self.terrain_runtime.world_size = world_size
        self.renderer.sync_viewport(
            world_size=world_size,
            config=self.config,
            camera=self.camera,
        )
        ground = self.render_resources.ground
        if ground is not None:
            side = max(0, int(world_size))
            ground.width = side
            ground.height = side

    def reset(
        self,
        *,
        seed: int = 0xBEEF,
        player_count: int = 1,
        spawn_pos: Vec2 | None = None,
    ) -> None:
        self._sync_world_size_ownership()
        self.sim_world.demo_mode_active = bool(self.demo_mode_active)
        self.sim_world.hardcore = bool(self.hardcore)
        self.sim_world.quest_fail_retry_count = int(self.quest_fail_retry_count)
        self.sim_world.preserve_bugs = bool(self.preserve_bugs)
        self.sim_world.reset(
            seed=int(seed),
            player_count=int(player_count),
            spawn_pos=spawn_pos,
        )
        self.render_resources.clear_pending_terrain_fx()
        self.camera = Vec2(-1.0, -1.0)
        self.renderer.sync_viewport(
            world_size=self.world_size,
            config=self.config,
            camera=self.camera,
        )
        if self.render_resources.ground is not None:
            terrain_seed = self.sim_world.state.rng.state
            self.terrain_runtime.schedule_from_rng_seed(seed=terrain_seed)

    def open_runtime(self) -> None:
        self.render_resources.config = self.config
        self.render_resources.open(terrain_seed=self.sim_world.state.rng.state)

    def close_runtime(self) -> None:
        self.render_resources.close()
        self.sim_world.close_session()

    def sync_audio_bridge_state(self) -> None:
        self.audio_bridge.sync(
            audio=self.audio,
            audio_rng=self.audio_rng,
            demo_mode_active=bool(self.demo_mode_active),
        )

    def update_camera(self, dt: float) -> None:
        _ = dt
        if not self.sim_world.players:
            return

        screen_size = viewport.camera_screen_size(
            world_size=self.world_size,
            config=self.config,
            runtime_w=float(rl.get_screen_width()),
            runtime_h=float(rl.get_screen_height()),
        )
        alive = [player for player in self.sim_world.players if player.health > 0.0]
        if alive:
            inv_alive = 1.0 / float(len(alive))
            focus = Vec2(
                sum(player.pos.x for player in alive) * inv_alive,
                sum(player.pos.y for player in alive) * inv_alive,
            )
            camera = screen_size * 0.5 - focus
        else:
            camera = self.camera

        camera = camera + self.sim_world.state.camera_shake_offset
        self.camera = viewport.clamp_camera(
            world_size=self.world_size,
            camera=camera,
            screen_size=screen_size,
        )
        self.renderer.sync_viewport(
            world_size=self.world_size,
            config=self.config,
            camera=self.camera,
        )

    def draw(
        self,
        *,
        draw_aim_indicators: bool = True,
        entity_alpha: float = 1.0,
    ) -> None:
        self.renderer.draw(
            render_frame=self.build_render_frame(),
            draw_aim_indicators=draw_aim_indicators,
            entity_alpha=entity_alpha,
        )

    def build_render_frame(self) -> RenderFrame:
        return self.render_resources.build_render_frame(
            state=self.sim_world.state,
            players=self.sim_world.players,
            creatures=self.sim_world.creatures,
            camera=self.camera,
            demo_mode_active=bool(self.demo_mode_active),
            elapsed_ms=float(self.sim_world.presentation_elapsed_ms),
            bonus_anim_phase=float(self.sim_world.bonus_anim_phase),
            lan_player_rings_enabled=bool(self.lan_player_rings_enabled),
            lan_local_aim_indicators_only=bool(self.lan_local_aim_indicators_only),
            lan_local_player_slot_index=int(self.lan_local_player_slot_index),
            rtx_mode=self.rtx_mode,
        )



from __future__ import annotations

from enum import IntEnum, IntFlag

Tint = tuple[float | None, float | None, float | None, float | None]
TintRGBA = tuple[float, float, float, float]

# Heading sentinel that forces randomized heading in `creature_spawn_template`.
RANDOM_HEADING_SENTINEL = -100.0


class CreatureTypeId(IntEnum):
    ZOMBIE = 0
    LIZARD = 1
    ALIEN = 2
    SPIDER_SP1 = 3
    SPIDER_SP2 = 4
    TROOPER = 5


class CreatureAiMode(IntEnum):
    ORBIT_PLAYER = 0
    ORBIT_PLAYER_TIGHT = 1
    CHASE_PLAYER = 2
    FOLLOW_LINK = 3
    LINK_GUARD = 4
    FOLLOW_LINK_TETHERED = 5
    ORBIT_LINK = 6
    HOLD_TIMER = 7
    ORBIT_PLAYER_WIDE = 8


class CreatureFlags(IntFlag):
    SELF_DAMAGE_TICK = 0x01  # periodic self-damage tick (dt * 60)
    SELF_DAMAGE_TICK_STRONG = 0x02  # stronger periodic self-damage tick (dt * 180)
    ANIM_PING_PONG = 0x04  # short ping-pong strip
    SPLIT_ON_DEATH = 0x08  # split-on-death behavior
    RANGED_ATTACK_SHOCK = 0x10  # ranged attack using projectile type 9
    ANIM_LONG_STRIP = 0x40  # force long animation strip
    AI7_LINK_TIMER = 0x80  # uses link index as timer for AI mode 7
    RANGED_ATTACK_VARIANT = 0x100  # ranged attack using orbit_radius as projectile type
    BONUS_ON_DEATH = 0x400  # spawns bonus on death


# Same bit as `ANIM_PING_PONG`; spawn logic reuses it to mean that `link_index`
# is interpreted as a spawn slot index in template/runtime paths.
HAS_SPAWN_SLOT_FLAG = CreatureFlags.ANIM_PING_PONG


class SpawnId(IntEnum):
    ZOMBIE_BOSS_SPAWNER_00 = 0x00
    SPIDER_SP2_SPLITTER_01 = 0x01
    UNUSED_02 = 0x02
    SPIDER_SP1_RANDOM_03 = 0x03
    LIZARD_RANDOM_04 = 0x04
    SPIDER_SP2_RANDOM_05 = 0x05
    ALIEN_RANDOM_06 = 0x06

    ALIEN_SPAWNER_CHILD_1D_FAST_07 = 0x07
    ALIEN_SPAWNER_CHILD_1D_SLOW_08 = 0x08
    ALIEN_SPAWNER_CHILD_1D_LIMITED_09 = 0x09
    ALIEN_SPAWNER_CHILD_32_SLOW_0A = 0x0A
    ALIEN_SPAWNER_CHILD_3C_SLOW_0B = 0x0B
    ALIEN_SPAWNER_CHILD_31_FAST_0C = 0x0C
    ALIEN_SPAWNER_CHILD_31_SLOW_0D = 0x0D
    ALIEN_SPAWNER_RING_24_0E = 0x0E
    ALIEN_CONST_BROWN_TRANSPARENT_0F = 0x0F
    ALIEN_SPAWNER_CHILD_32_FAST_10 = 0x10

    FORMATION_CHAIN_LIZARD_4_11 = 0x11
    FORMATION_RING_ALIEN_8_12 = 0x12
    FORMATION_CHAIN_ALIEN_10_13 = 0x13
    FORMATION_GRID_ALIEN_GREEN_14 = 0x14
    FORMATION_GRID_ALIEN_WHITE_15 = 0x15
    FORMATION_GRID_LIZARD_WHITE_16 = 0x16
    FORMATION_GRID_SPIDER_SP1_WHITE_17 = 0x17
    FORMATION_GRID_ALIEN_BRONZE_18 = 0x18
    FORMATION_RING_ALIEN_5_19 = 0x19

    AI1_ALIEN_BLUE_TINT_1A = 0x1A
    AI1_SPIDER_SP1_BLUE_TINT_1B = 0x1B
    AI1_LIZARD_BLUE_TINT_1C = 0x1C

    ALIEN_RANDOM_1D = 0x1D
    ALIEN_RANDOM_1E = 0x1E
    ALIEN_RANDOM_1F = 0x1F
    ALIEN_RANDOM_GREEN_20 = 0x20

    ALIEN_CONST_PURPLE_GHOST_21 = 0x21
    ALIEN_CONST_GREEN_GHOST_22 = 0x22
    ALIEN_CONST_GREEN_GHOST_SMALL_23 = 0x23
    ALIEN_CONST_GREEN_24 = 0x24
    ALIEN_CONST_GREEN_SMALL_25 = 0x25
    ALIEN_CONST_PALE_GREEN_26 = 0x26
    ALIEN_CONST_WEAPON_BONUS_27 = 0x27
    ALIEN_CONST_PURPLE_28 = 0x28
    ALIEN_CONST_GREY_BRUTE_29 = 0x29
    ALIEN_CONST_GREY_FAST_2A = 0x2A
    ALIEN_CONST_RED_FAST_2B = 0x2B
    ALIEN_CONST_RED_BOSS_2C = 0x2C
    ALIEN_CONST_CYAN_AI2_2D = 0x2D

    LIZARD_RANDOM_2E = 0x2E
    LIZARD_CONST_GREY_2F = 0x2F
    LIZARD_CONST_YELLOW_BOSS_30 = 0x30
    LIZARD_RANDOM_31 = 0x31

    SPIDER_SP1_RANDOM_32 = 0x32
    SPIDER_SP1_RANDOM_RED_33 = 0x33
    SPIDER_SP1_RANDOM_GREEN_34 = 0x34
    SPIDER_SP2_RANDOM_35 = 0x35

    ALIEN_AI7_ORBITER_36 = 0x36
    SPIDER_SP2_RANGED_VARIANT_37 = 0x37
    SPIDER_SP1_AI7_TIMER_38 = 0x38
    SPIDER_SP1_AI7_TIMER_WEAK_39 = 0x39

    SPIDER_SP1_CONST_SHOCK_BOSS_3A = 0x3A
    SPIDER_SP1_CONST_RED_BOSS_3B = 0x3B
    SPIDER_SP1_CONST_RANGED_VARIANT_3C = 0x3C
    SPIDER_SP1_RANDOM_3D = 0x3D
    SPIDER_SP1_CONST_WHITE_FAST_3E = 0x3E
    SPIDER_SP1_CONST_BROWN_SMALL_3F = 0x3F
    SPIDER_SP1_CONST_BLUE_40 = 0x40

    ZOMBIE_RANDOM_41 = 0x41
    ZOMBIE_CONST_GREY_42 = 0x42
    ZOMBIE_CONST_GREEN_BRUTE_43 = 0x43



from __future__ import annotations

from typing import TYPE_CHECKING

from grim.assets import RuntimeResources, TextureId
from grim.fonts.small import SmallFontData, draw_small_text, measure_small_text_width
from grim.geom import Vec2
from grim.raylib_api import rl

from ...game_modes import GameMode
from ..high_scores_layout import (
    HS_LOCAL_CLOCK_X,
    HS_LOCAL_CLOCK_Y,
    HS_LOCAL_DATE_X,
    HS_LOCAL_DATE_Y,
    HS_LOCAL_FRAGS_X,
    HS_LOCAL_FRAGS_Y,
    HS_LOCAL_HIT_X,
    HS_LOCAL_HIT_Y,
    HS_LOCAL_LABEL_X,
    HS_LOCAL_LABEL_Y,
    HS_LOCAL_NAME_X,
    HS_LOCAL_NAME_Y,
    HS_LOCAL_RANK_X,
    HS_LOCAL_RANK_Y,
    HS_LOCAL_SCORE_LABEL_X,
    HS_LOCAL_SCORE_LABEL_Y,
    HS_LOCAL_SCORE_VALUE_X,
    HS_LOCAL_SCORE_VALUE_Y,
    HS_LOCAL_TIME_LABEL_X,
    HS_LOCAL_TIME_LABEL_Y,
    HS_LOCAL_TIME_VALUE_X,
    HS_LOCAL_TIME_VALUE_Y,
    HS_LOCAL_WEAPON_Y,
    HS_LOCAL_WICON_X,
    HS_LOCAL_WICON_Y,
    HS_RIGHT_CHECK_X,
    HS_RIGHT_CHECK_Y,
    HS_RIGHT_GAME_MODE_DROP_X,
    HS_RIGHT_GAME_MODE_DROP_Y,
    HS_RIGHT_GAME_MODE_VALUE_X,
    HS_RIGHT_GAME_MODE_VALUE_Y,
    HS_RIGHT_GAME_MODE_WIDGET_W,
    HS_RIGHT_GAME_MODE_WIDGET_X,
    HS_RIGHT_GAME_MODE_WIDGET_Y,
    HS_RIGHT_GAME_MODE_X,
    HS_RIGHT_GAME_MODE_Y,
    HS_RIGHT_NUMBER_PLAYERS_X,
    HS_RIGHT_NUMBER_PLAYERS_Y,
    HS_RIGHT_PLAYER_COUNT_DROP_X,
    HS_RIGHT_PLAYER_COUNT_DROP_Y,
    HS_RIGHT_PLAYER_COUNT_VALUE_X,
    HS_RIGHT_PLAYER_COUNT_VALUE_Y,
    HS_RIGHT_PLAYER_COUNT_WIDGET_W,
    HS_RIGHT_PLAYER_COUNT_WIDGET_X,
    HS_RIGHT_PLAYER_COUNT_WIDGET_Y,
    HS_RIGHT_SCORE_LIST_DROP_X,
    HS_RIGHT_SCORE_LIST_DROP_Y,
    HS_RIGHT_SCORE_LIST_VALUE_X,
    HS_RIGHT_SCORE_LIST_VALUE_Y,
    HS_RIGHT_SCORE_LIST_WIDGET_W,
    HS_RIGHT_SCORE_LIST_WIDGET_X,
    HS_RIGHT_SCORE_LIST_WIDGET_Y,
    HS_RIGHT_SCORE_LIST_X,
    HS_RIGHT_SCORE_LIST_Y,
    HS_RIGHT_SHOW_INTERNET_X,
    HS_RIGHT_SHOW_INTERNET_Y,
    HS_RIGHT_SHOW_SCORES_DROP_X,
    HS_RIGHT_SHOW_SCORES_DROP_Y,
    HS_RIGHT_SHOW_SCORES_VALUE_X,
    HS_RIGHT_SHOW_SCORES_VALUE_Y,
    HS_RIGHT_SHOW_SCORES_WIDGET_W,
    HS_RIGHT_SHOW_SCORES_WIDGET_X,
    HS_RIGHT_SHOW_SCORES_WIDGET_Y,
    HS_RIGHT_SHOW_SCORES_X,
    HS_RIGHT_SHOW_SCORES_Y,
    hs_right_local_card_x_shift,
    hs_right_options_x_shift,
)
from ..panels.hit_test import mouse_inside_rect_with_padding
from .shared import format_elapsed_mm_ss, format_score_date, ordinal

if TYPE_CHECKING:
    from .view import HighScoresView


def _saved_score_names(view: "HighScoresView") -> list[str]:
    return list(view.state.config.profile.saved_name_labels())


def _draw_dropdown(
    *,
    resources: RuntimeResources,
    font: SmallFontData,
    widget_pos: Vec2,
    widget_w: float,
    items: list[str] | tuple[str, ...],
    selected_index: int,
    value_pos: Vec2,
    arrow_pos: Vec2,
    is_open: bool,
    enabled: bool,
    scale: float,
) -> None:
    item_count = max(0, len(items))
    header_h = 16.0 * scale
    row_h = 16.0 * scale
    full_h = (float(item_count) * 16.0 + 24.0) * scale
    rows_y0 = widget_pos.y + 17.0 * scale

    mouse = rl.get_mouse_position()
    hovered_header = bool(enabled) and mouse_inside_rect_with_padding(
        mouse,
        pos=widget_pos,
        width=widget_w,
        height=14.0 * scale,
    )

    widget_h = full_h if is_open else header_h
    rl.draw_rectangle(int(widget_pos.x), int(widget_pos.y), int(widget_w), int(widget_h), rl.WHITE)
    rl.draw_rectangle(
        int(widget_pos.x) + 1,
        int(widget_pos.y) + 1,
        max(0, int(widget_w) - 2),
        max(0, int(widget_h) - 2),
        rl.BLACK,
    )

    if (is_open or hovered_header) and enabled:
        line_h = max(1, int(1.0 * scale))
        rl.draw_rectangle(
            int(widget_pos.x),
            int(widget_pos.y + 15.0 * scale),
            int(widget_w),
            line_h,
            rl.Color(255, 255, 255, 128),
        )

    arrow_tex = (
        resources.texture(TextureId.UI_DROP_ON)
        if ((is_open or hovered_header) and enabled)
        else resources.texture(TextureId.UI_DROP_OFF)
    )
    arrow_w = float(arrow_tex.width) * scale
    arrow_h = float(arrow_tex.height) * scale
    rl.draw_texture_pro(
        arrow_tex,
        rl.Rectangle(0.0, 0.0, float(arrow_tex.width), float(arrow_tex.height)),
        rl.Rectangle(arrow_pos.x, arrow_pos.y, arrow_w, arrow_h),
        rl.Vector2(0.0, 0.0),
        0.0,
        rl.WHITE,
    )

    if item_count <= 0:
        return

    selected_index = max(0, min(item_count - 1, int(selected_index)))
    header_alpha = 242 if ((is_open or hovered_header) and enabled) else 191
    draw_small_text(font, str(items[selected_index]), value_pos, rl.Color(255, 255, 255, header_alpha))

    if not is_open:
        return

    for idx, label in enumerate(items):
        item_y = rows_y0 + row_h * float(idx)
        hovered = bool(enabled) and mouse_inside_rect_with_padding(
            mouse,
            pos=Vec2(widget_pos.x, item_y),
            width=widget_w,
            height=14.0 * scale,
        )
        alpha = 153
        if hovered:
            alpha = 242
        if idx == selected_index:
            alpha = max(alpha, 245)
        draw_small_text(font, str(label), Vec2(value_pos.x, item_y), rl.Color(255, 255, 255, alpha))


def draw_right_panel(
    view: "HighScoresView",
    *,
    resources: RuntimeResources,
    font: SmallFontData,
    right_top_left: Vec2,
    scale: float,
    highlight_rank: int | None,
) -> None:
    if highlight_rank is None:
        _draw_right_panel_quest_options(
            view,
            resources=resources,
            font=font,
            right_top_left=right_top_left,
            scale=scale,
        )
        return
    _draw_right_panel_local_score(
        view,
        resources=resources,
        font=font,
        right_top_left=right_top_left,
        scale=scale,
        highlight_rank=highlight_rank,
    )


def _draw_right_panel_quest_options(
    view: "HighScoresView",
    *,
    resources: RuntimeResources,
    font: SmallFontData,
    right_top_left: Vec2,
    scale: float,
) -> None:
    options_shift_x = hs_right_options_x_shift(float(view.state.config.display.width))
    options_top_left = right_top_left + Vec2(options_shift_x * scale, 0.0)
    text_color = rl.Color(255, 255, 255, int(255 * 0.8))

    # Checkbox: "Show internet scores"
    check_tex = (
        resources.texture(TextureId.UI_CHECK_ON)
        if view.state.config.profile.show_internet_scores
        else resources.texture(TextureId.UI_CHECK_OFF)
    )
    check_w = float(check_tex.width) * scale
    check_h = float(check_tex.height) * scale
    rl.draw_texture_pro(
        check_tex,
        rl.Rectangle(0.0, 0.0, float(check_tex.width), float(check_tex.height)),
        rl.Rectangle(
            options_top_left.x + HS_RIGHT_CHECK_X * scale,
            options_top_left.y + HS_RIGHT_CHECK_Y * scale,
            check_w,
            check_h,
        ),
        rl.Vector2(0.0, 0.0),
        0.0,
        rl.WHITE,
    )
    draw_small_text(font, "Show internet scores", options_top_left + Vec2(HS_RIGHT_SHOW_INTERNET_X * scale, HS_RIGHT_SHOW_INTERNET_Y * scale), text_color)
    draw_small_text(font, "Number of players", options_top_left + Vec2(HS_RIGHT_NUMBER_PLAYERS_X * scale, HS_RIGHT_NUMBER_PLAYERS_Y * scale), text_color)
    draw_small_text(font, "Game mode", options_top_left + Vec2(HS_RIGHT_GAME_MODE_X * scale, HS_RIGHT_GAME_MODE_Y * scale), text_color)
    draw_small_text(font, "Show scores:", options_top_left + Vec2(HS_RIGHT_SHOW_SCORES_X * scale, HS_RIGHT_SHOW_SCORES_Y * scale), text_color)
    draw_small_text(font, "Selected score list:", options_top_left + Vec2(HS_RIGHT_SCORE_LIST_X * scale, HS_RIGHT_SCORE_LIST_Y * scale), text_color)

    # Dropdown widgets (state_14 quest variant).
    show_scores_items = ("Best of all time", "Best of month", "Best of week", "Best of day")
    player_items = ("1 player", "2 players", "3 players", "4 players")
    mode_items: list[tuple[str, int]] = [("Quests", 3), ("Rush", 2), ("Survival", 1)]
    if int(view.state.status.quest_unlock_index) >= 0x28:
        mode_items.append(("Typ'o'Shooter", 4))
    names = _saved_score_names(view)

    player_count = max(1, min(4, view.state.config.gameplay.player_count))
    player_selected = player_count - 1
    show_scores_selected = max(0, min(len(show_scores_items) - 1, int(view.state.config.profile.score_date_mode)))
    mode_id = view.state.config.gameplay.mode
    mode_selected = 0
    for idx, (_label, _id) in enumerate(mode_items):
        if int(_id) == int(mode_id):
            mode_selected = idx
            break
    name_selected = max(0, min(len(names) - 1, int(view.state.config.profile.selected_saved_name_slot)))

    dropdowns = (
        (
            view._player_count_open,
            Vec2(HS_RIGHT_PLAYER_COUNT_WIDGET_X, HS_RIGHT_PLAYER_COUNT_WIDGET_Y),
            float(HS_RIGHT_PLAYER_COUNT_WIDGET_W),
            list(player_items),
            player_selected,
            Vec2(HS_RIGHT_PLAYER_COUNT_VALUE_X, HS_RIGHT_PLAYER_COUNT_VALUE_Y),
            Vec2(HS_RIGHT_PLAYER_COUNT_DROP_X, HS_RIGHT_PLAYER_COUNT_DROP_Y),
            not (view._game_mode_open or view._show_scores_open or view._score_list_open),
        ),
        (
            view._game_mode_open,
            Vec2(HS_RIGHT_GAME_MODE_WIDGET_X, HS_RIGHT_GAME_MODE_WIDGET_Y),
            float(HS_RIGHT_GAME_MODE_WIDGET_W),
            [label for label, _id in mode_items],
            mode_selected,
            Vec2(HS_RIGHT_GAME_MODE_VALUE_X, HS_RIGHT_GAME_MODE_VALUE_Y),
            Vec2(HS_RIGHT_GAME_MODE_DROP_X, HS_RIGHT_GAME_MODE_DROP_Y),
            not (view._player_count_open or view._show_scores_open or view._score_list_open),
        ),
        (
            view._show_scores_open,
            Vec2(HS_RIGHT_SHOW_SCORES_WIDGET_X, HS_RIGHT_SHOW_SCORES_WIDGET_Y),
            float(HS_RIGHT_SHOW_SCORES_WIDGET_W),
            list(show_scores_items),
            show_scores_selected,
            Vec2(HS_RIGHT_SHOW_SCORES_VALUE_X, HS_RIGHT_SHOW_SCORES_VALUE_Y),
            Vec2(HS_RIGHT_SHOW_SCORES_DROP_X, HS_RIGHT_SHOW_SCORES_DROP_Y),
            not (view._player_count_open or view._game_mode_open or view._score_list_open),
        ),
        (
            view._score_list_open,
            Vec2(HS_RIGHT_SCORE_LIST_WIDGET_X, HS_RIGHT_SCORE_LIST_WIDGET_Y),
            float(HS_RIGHT_SCORE_LIST_WIDGET_W),
            names,
            name_selected,
            Vec2(HS_RIGHT_SCORE_LIST_VALUE_X, HS_RIGHT_SCORE_LIST_VALUE_Y),
            Vec2(HS_RIGHT_SCORE_LIST_DROP_X, HS_RIGHT_SCORE_LIST_DROP_Y),
            not (view._player_count_open or view._game_mode_open or view._show_scores_open),
        ),
    )
    # Active list must render last so overlapping widgets don't occlude open options.
    for is_open, widget_offset, widget_w, items, selected_index, value_offset, arrow_offset, enabled in dropdowns:
        if is_open:
            continue
        _draw_dropdown(
            resources=resources,
            font=font,
            widget_pos=options_top_left + widget_offset * scale,
            widget_w=widget_w * scale,
            items=items,
            selected_index=selected_index,
            value_pos=options_top_left + value_offset * scale,
            arrow_pos=options_top_left + arrow_offset * scale,
            is_open=is_open,
            enabled=bool(enabled),
            scale=scale,
        )
    for is_open, widget_offset, widget_w, items, selected_index, value_offset, arrow_offset, enabled in dropdowns:
        if not is_open:
            continue
        _draw_dropdown(
            resources=resources,
            font=font,
            widget_pos=options_top_left + widget_offset * scale,
            widget_w=widget_w * scale,
            items=items,
            selected_index=selected_index,
            value_pos=options_top_left + value_offset * scale,
            arrow_pos=options_top_left + arrow_offset * scale,
            is_open=is_open,
            enabled=bool(enabled),
            scale=scale,
        )


def _draw_right_panel_local_score(
    view: "HighScoresView",
    *,
    resources: RuntimeResources,
    font: SmallFontData,
    right_top_left: Vec2,
    scale: float,
    highlight_rank: int | None,
) -> None:
    local_shift_x = hs_right_local_card_x_shift(float(view.state.config.display.width))
    card_top_left = right_top_left + Vec2(local_shift_x * scale, 0.0)
    if not view._records:
        return
    idx = int(highlight_rank) if highlight_rank is not None else int(view._scroll_index)
    if idx < 0:
        idx = 0
    if idx >= len(view._records):
        idx = len(view._records) - 1
    entry = view._records[idx]

    text_color = rl.Color(int(255 * 0.9), int(255 * 0.9), int(255 * 0.9), int(255 * 0.8))
    value_color = rl.Color(int(255 * 0.9), int(255 * 0.9), 255, 255)
    game_time_color = rl.Color(255, 255, 255, int(255 * 0.8))
    lower_section_color = rl.Color(int(255 * 0.9), int(255 * 0.9), int(255 * 0.9), int(255 * 0.7))
    separator_color = rl.Color(149, 175, 198, int(255 * 0.7))

    name = str(entry.name())
    if not name:
        name = "???"
    draw_small_text(font, name, card_top_left + Vec2(HS_LOCAL_NAME_X * scale, HS_LOCAL_NAME_Y * scale), text_color)
    draw_small_text(font, "Local score", card_top_left + Vec2(HS_LOCAL_LABEL_X * scale, HS_LOCAL_LABEL_Y * scale), text_color)
    rl.draw_line(
        int(card_top_left.x + 78.0 * scale),
        int(card_top_left.y + 57.0 * scale),
        int(card_top_left.x + 117.0 * scale),
        int(card_top_left.y + 57.0 * scale),
        separator_color,
    )

    date_text = format_score_date(entry)
    if date_text:
        draw_small_text(font, date_text, card_top_left + Vec2(HS_LOCAL_DATE_X * scale, HS_LOCAL_DATE_Y * scale), text_color)
    rl.draw_line(
        int(card_top_left.x + 74.0 * scale),
        int(card_top_left.y + 72.0 * scale),
        int(card_top_left.x + 266.0 * scale),
        int(card_top_left.y + 72.0 * scale),
        separator_color,
    )

    draw_small_text(font, "Score", card_top_left + Vec2(HS_LOCAL_SCORE_LABEL_X * scale, HS_LOCAL_SCORE_LABEL_Y * scale), text_color)

    mode_raw = int(entry.game_mode_id)
    try:
        mode_id = GameMode(mode_raw)
    except ValueError:
        mode_id = GameMode.DEMO
    elapsed_ms = int(entry.survival_elapsed_ms)
    score_xp = int(entry.score_xp)
    match mode_id:
        case GameMode.QUESTS:
            time_label = "Experience"
        case _:
            time_label = "Game time"

    draw_small_text(font, time_label, card_top_left + Vec2(HS_LOCAL_TIME_LABEL_X * scale, HS_LOCAL_TIME_LABEL_Y * scale), game_time_color)
    rl.draw_line(
        int(card_top_left.x + 170.0 * scale),
        int(card_top_left.y + 90.0 * scale),
        int(card_top_left.x + 170.0 * scale),
        int(card_top_left.y + 138.0 * scale),
        separator_color,
    )

    # Native highscore card:
    # - Rush/Quest: score is survival time in seconds (ms * 0.001), rendered with 2 decimals.
    # - Others: score is XP (u32).
    score_value_pos = Vec2(HS_LOCAL_SCORE_VALUE_X * scale, HS_LOCAL_SCORE_VALUE_Y * scale)
    match mode_id:
        case GameMode.RUSH | GameMode.QUESTS:
            score_value = f"{max(0, elapsed_ms) * 0.001:.2f} secs"
            # Quest/Rush scores are variable-width second labels ("%.2f secs") and are
            # centered in the left score column in native.
            score_label_w = measure_small_text_width(font, "Score")
            score_value_w = measure_small_text_width(font, score_value)
            score_col_center_x = HS_LOCAL_SCORE_LABEL_X * scale + score_label_w * 0.5
            score_value_pos = Vec2(score_col_center_x - score_value_w * 0.5, HS_LOCAL_SCORE_VALUE_Y * scale)
        case _:
            score_value = f"{score_xp}"
    draw_small_text(font, score_value, card_top_left + score_value_pos, value_color)

    match mode_id:
        case GameMode.QUESTS:
            draw_small_text(font, f"{score_xp}", card_top_left + Vec2(HS_LOCAL_TIME_VALUE_X * scale, HS_LOCAL_TIME_VALUE_Y * scale), game_time_color)
        case _:
            _draw_clock_gauge(
                resources=resources,
                elapsed_ms=elapsed_ms,
                pos=card_top_left + Vec2(HS_LOCAL_CLOCK_X * scale, HS_LOCAL_CLOCK_Y * scale),
                scale=scale,
            )
            draw_small_text(font, format_elapsed_mm_ss(elapsed_ms), card_top_left + Vec2(HS_LOCAL_TIME_VALUE_X * scale, HS_LOCAL_TIME_VALUE_Y * scale), game_time_color)

    draw_small_text(font, f"Rank: {ordinal(idx + 1)}", card_top_left + Vec2(HS_LOCAL_RANK_X * scale, HS_LOCAL_RANK_Y * scale), text_color)

    frags = int(entry.creature_kill_count)

    shots_fired = int(entry.shots_fired)
    shots_hit = int(entry.shots_hit)
    hit_pct = 0
    if shots_fired > 0:
        hit_pct = int((shots_hit * 100) // shots_fired)
    rl.draw_line(
        int(card_top_left.x + 74.0 * scale),
        int(card_top_left.y + 142.0 * scale),
        int(card_top_left.x + 266.0 * scale),
        int(card_top_left.y + 142.0 * scale),
        separator_color,
    )

    weapon_id = entry.most_used_weapon_id
    weapon_name, icon_index = _weapon_label_and_icon(view, weapon_id)
    if icon_index is not None:
        _draw_wicon(
            resources=resources,
            icon_index=icon_index,
            pos=card_top_left + Vec2(HS_LOCAL_WICON_X * scale, HS_LOCAL_WICON_Y * scale),
            scale=scale,
        )
    weapon_name_x = HS_LOCAL_WICON_X * scale + max(
        0.0,
        32.0 * scale - measure_small_text_width(font, weapon_name) * 0.5,
    )
    draw_small_text(font, weapon_name, card_top_left + Vec2(weapon_name_x, HS_LOCAL_WEAPON_Y * scale), lower_section_color)
    draw_small_text(font, f"Frags: {frags}", card_top_left + Vec2(HS_LOCAL_FRAGS_X * scale, HS_LOCAL_FRAGS_Y * scale), lower_section_color)
    draw_small_text(font, f"Hit %: {hit_pct}%", card_top_left + Vec2(HS_LOCAL_HIT_X * scale, HS_LOCAL_HIT_Y * scale), lower_section_color)
    rl.draw_line(
        int(card_top_left.x + 74.0 * scale),
        int(card_top_left.y + 194.0 * scale),
        int(card_top_left.x + 266.0 * scale),
        int(card_top_left.y + 194.0 * scale),
        separator_color,
    )


def _draw_clock_gauge(
    *,
    resources: RuntimeResources,
    elapsed_ms: int,
    pos: Vec2,
    scale: float,
) -> None:
    table_tex = resources.texture(TextureId.UI_CLOCK_TABLE)
    pointer_tex = resources.texture(TextureId.UI_CLOCK_POINTER)
    draw_w = 32.0 * scale
    draw_h = 32.0 * scale
    dst = rl.Rectangle(pos.x, pos.y, draw_w, draw_h)
    src_table = rl.Rectangle(0.0, 0.0, float(table_tex.width), float(table_tex.height))
    src_pointer = rl.Rectangle(0.0, 0.0, float(pointer_tex.width), float(pointer_tex.height))
    rl.draw_texture_pro(
        table_tex,
        src_table,
        dst,
        rl.Vector2(0.0, 0.0),
        0.0,
        rl.WHITE,
    )
    seconds = max(0, int(elapsed_ms) // 1000)
    rotation_deg = float(seconds) * 6.0
    center = Vec2(pos.x + draw_w * 0.5, pos.y + draw_h * 0.5)
    rl.draw_texture_pro(
        pointer_tex,
        src_pointer,
        rl.Rectangle(center.x, center.y, draw_w, draw_h),
        rl.Vector2(draw_w * 0.5, draw_h * 0.5),
        rotation_deg,
        rl.WHITE,
    )


def _draw_wicon(
    *,
    resources: RuntimeResources,
    icon_index: int,
    pos: Vec2,
    scale: float,
) -> None:
    tex = resources.texture(TextureId.UI_WICONS)
    idx = int(icon_index)
    if idx < 0 or idx > 31:
        return
    grid = 8
    cell_w = float(tex.width) / float(grid)
    cell_h = float(tex.height) / float(grid)
    frame = idx * 2
    src_x = float(frame % grid) * cell_w
    src_y = float(frame // grid) * cell_h
    icon_w = cell_w * 2.0
    icon_h = cell_h
    rl.draw_texture_pro(
        tex,
        rl.Rectangle(src_x, src_y, icon_w, icon_h),
        rl.Rectangle(pos.x, pos.y, icon_w * scale, icon_h * scale),
        rl.Vector2(0.0, 0.0),
        0.0,
        rl.WHITE,
    )


def _weapon_label_and_icon(view: "HighScoresView", weapon_id: int) -> tuple[str, int | None]:
    from ...weapons import WEAPON_BY_ID, WeaponId, weapon_display_name

    weapon = WEAPON_BY_ID[WeaponId(weapon_id)]
    name = weapon_display_name(weapon.weapon_id, preserve_bugs=bool(view.state.preserve_bugs))
    return name, weapon.icon_index


__all__ = ["draw_right_panel"]



from __future__ import annotations

from collections.abc import Sequence

import msgspec

from grim.geom import Vec2

from .names import CreatureNameTable
from .typing import TypingBuffer


class TypoState(msgspec.Struct):
    typing: TypingBuffer = msgspec.field(default_factory=TypingBuffer)
    names: CreatureNameTable = msgspec.field(default_factory=lambda: CreatureNameTable.sized(0))
    spawn_cooldown_ms: int = 0
    dictionary_words: tuple[str, ...] = ()
    highscore_names: tuple[str, ...] = ()
    pending_fire_target: Vec2 | None = None
    pending_reload: bool = False


def reset_typo_state(
    typo: TypoState,
    *,
    creature_capacity: int,
    dictionary_words: Sequence[str] = (),
    highscore_names: Sequence[str] = (),
) -> None:
    typo.typing = TypingBuffer()
    typo.names = CreatureNameTable.sized(int(creature_capacity))
    typo.spawn_cooldown_ms = 0
    typo.dictionary_words = tuple(str(word) for word in dictionary_words)
    typo.highscore_names = tuple(str(name) for name in highscore_names)
    typo.pending_fire_target = None
    typo.pending_reload = False


def typo_shot_counts(typo: TypoState) -> tuple[int, int]:
    return int(typo.typing.submit_count), int(typo.typing.match_count)



from __future__ import annotations

from pathlib import Path

from .channel_helpers import (
    entity_samples_channel,
    rng_stream_channel_required,
    sim_state_channel,
    timing_samples_channel_required,
)
from .payloads import BuiltinObject, to_builtin_object
from .schema import TRACE_REQUIRED_CHANNELS
from .trace import TraceReader


def summarize_trace_health(
    trace_path: Path,
    *,
    tick_start: int | None = None,
    tick_end: int | None = None,
) -> BuiltinObject:
    path = Path(trace_path)
    with TraceReader(path) as trace:
        channels_present: dict[str, int] = {}
        ticks_total = 0
        ticks_with_dt = 0
        rng_stream_rows = 0
        sim_state_rows = 0
        sample_creature_rows = 0
        sample_projectile_rows = 0
        sample_secondary_rows = 0
        sample_bonus_rows = 0
        timing_samples_rows = 0

        for tick in trace.iter_ticks(tick_start=tick_start, tick_end=tick_end):
            ticks_total += 1
            ticks_with_dt += 1
            for channel_name in TRACE_REQUIRED_CHANNELS:
                channels_present[channel_name] = int(channels_present.get(channel_name, 0)) + 1
                if channel_name == "rng_stream":
                    rng_stream_rows += len(rng_stream_channel_required(tick))
                elif channel_name == "timing_samples":
                    timing_samples_rows += len(timing_samples_channel_required(tick))
                elif channel_name == "sim_state":
                    if sim_state_channel(tick) is not None:
                        sim_state_rows += 1
                elif channel_name == "entity_samples":
                    samples = entity_samples_channel(tick)
                    if samples is not None:
                        sample_creature_rows += len(samples.creatures)
                        sample_projectile_rows += len(samples.projectiles)
                        sample_secondary_rows += len(samples.secondary_projectiles)
                        sample_bonus_rows += len(samples.bonuses)

        issues: list[str] = []
        if ticks_total == 0:
            issues.append("trace window has no ticks")
        for required_channel in TRACE_REQUIRED_CHANNELS:
            if channels_present.get(str(required_channel), 0) <= 0:
                issues.append(f"{required_channel} channel missing")

        window_start = tick_start
        window_end = tick_end
        meta_range = trace.meta.tick_range
        if ticks_total > 0:
            if window_start is None:
                window_start = int(meta_range.get("start_tick", -1))
            if window_end is None:
                window_end = int(meta_range.get("end_tick", -1))
        return to_builtin_object(
            {
                "trace_format_version": int(trace.meta.trace_format_version),
                "trace_schema_version": int(trace.meta.trace_schema_version),
                "tick_window": {
                    "requested_start": (None if tick_start is None else int(tick_start)),
                    "requested_end": (None if tick_end is None else int(tick_end)),
                    "actual_start": (None if window_start is None else int(window_start)),
                    "actual_end": (None if window_end is None else int(window_end)),
                    "ticks_in_window": int(ticks_total),
                },
                "channels_present": {str(key): int(value) for key, value in sorted(channels_present.items())},
                "metrics": {
                    "ticks_with_dt_ms_i32": int(ticks_with_dt),
                    "rng_stream_rows": int(rng_stream_rows),
                    "sim_state_rows": int(sim_state_rows),
                    "sample_creature_rows": int(sample_creature_rows),
                    "sample_projectile_rows": int(sample_projectile_rows),
                    "sample_secondary_projectile_rows": int(sample_secondary_rows),
                    "sample_bonus_rows": int(sample_bonus_rows),
                    "timing_samples_rows": int(timing_samples_rows),
                },
                "issues": issues,
                "ok_for_movement_root_cause": len(issues) == 0,
            },
            field="trace_health",
        )



from __future__ import annotations

from pathlib import Path

import msgspec

from .channel_compare import compare_entity_samples, compare_rng_stream, compare_sim_state, compare_timing_samples
from .channel_helpers import (
    checkpoint_channel_required,
    entity_samples_channel_required,
    rng_stream_channel_required,
    sim_state_channel_required,
    timing_samples_channel_required,
)
from .checkpoint_diff import checkpoint_deepdiff
from .payloads import BuiltinObject, to_builtin_object
from .schema import (
    TRACE_REQUIRED_CHANNELS,
    TickRecord,
)
from .trace import TraceError, TraceReader


class TraceMismatch(msgspec.Struct, frozen=True):
    kind: str
    tick_index: int
    detail: BuiltinObject | None = None


class TraceDiffReport(msgspec.Struct, frozen=True):
    ok: bool
    checked_count: int
    tick_start: int | None
    tick_end: int | None
    mismatch: TraceMismatch | None = None


class TraceBisectReport(msgspec.Struct, frozen=True):
    ok: bool
    first_bad_tick: int | None
    checked_count: int
    mismatch: TraceMismatch | None
    window_start: int | None = None
    window_end: int | None = None


class _TickPair(msgspec.Struct, frozen=True):
    tick_index: int
    expected_row: TickRecord | None
    actual_row: TickRecord | None


def _tick_mismatch_row(*, kind: str, channel: str, detail: BuiltinObject | None) -> BuiltinObject:
    return to_builtin_object(
        {
            "kind": str(kind),
            "channel": str(channel),
            "detail": (None if detail is None else to_builtin_object(detail, field=f"{channel}.detail")),
        },
        field=f"{channel}.mismatch",
    )


def _first_mismatch(
    *,
    pairs: list[_TickPair],
    tick_end: int | None = None,
) -> tuple[int, TraceMismatch | None]:
    checked_count = 0
    for pair in pairs:
        tick = pair.tick_index
        if tick_end is not None and tick > tick_end:
            break
        checked_count += 1
        if pair.expected_row is None or pair.actual_row is None:
            return (
                checked_count,
                TraceMismatch(
                    kind="missing_tick",
                    tick_index=tick,
                ),
            )
        tick_mismatches: list[BuiltinObject] = []

        expected = checkpoint_channel_required(pair.expected_row)
        actual = checkpoint_channel_required(pair.actual_row)
        exp_rng_stream = rng_stream_channel_required(pair.expected_row)
        act_rng_stream = rng_stream_channel_required(pair.actual_row)

        checkpoint_diff = checkpoint_deepdiff(expected, actual)
        if checkpoint_diff is not None:
            tick_mismatches.append(
                _tick_mismatch_row(
                    kind="checkpoint_field_mismatch",
                    channel="checkpoint",
                    detail={
                        "diff_count": int(checkpoint_diff.diff_count),
                        "payload": msgspec.to_builtins(checkpoint_diff.payload),
                        "pretty": str(checkpoint_diff.pretty),
                    },
                ),
            )

        rng_ok, rng_detail = compare_rng_stream(exp_rng_stream, act_rng_stream)
        if not rng_ok:
            tick_mismatches.append(
                _tick_mismatch_row(
                    kind="rng_stream_mismatch",
                    channel="rng_stream",
                    detail=rng_detail,
                ),
            )

        expected_sim_state = sim_state_channel_required(pair.expected_row)
        actual_sim_state = sim_state_channel_required(pair.actual_row)
        sim_ok, sim_detail = compare_sim_state(
            expected_sim_state,
            actual_sim_state,
        )
        if not sim_ok:
            tick_mismatches.append(
                _tick_mismatch_row(
                    kind="sim_state_mismatch",
                    channel="sim_state",
                    detail=sim_detail,
                ),
            )

        expected_entity_samples = entity_samples_channel_required(pair.expected_row)
        actual_entity_samples = entity_samples_channel_required(pair.actual_row)
        entities_ok, entities_detail = compare_entity_samples(
            expected_entity_samples,
            actual_entity_samples,
        )
        if not entities_ok:
            tick_mismatches.append(
                _tick_mismatch_row(
                    kind="entity_sample_mismatch",
                    channel="entity_samples",
                    detail=entities_detail,
                ),
            )

        expected_timing_samples = timing_samples_channel_required(pair.expected_row)
        actual_timing_samples = timing_samples_channel_required(pair.actual_row)
        timing_ok, timing_detail = compare_timing_samples(
            expected_timing_samples,
            actual_timing_samples,
        )
        if not timing_ok:
            tick_mismatches.append(
                _tick_mismatch_row(
                    kind="timing_samples_mismatch",
                    channel="timing_samples",
                    detail=timing_detail,
                ),
            )

        if tick_mismatches:
            return (
                checked_count,
                TraceMismatch(
                    kind="tick_mismatch",
                    tick_index=tick,
                    detail={
                        "mismatch_count": len(tick_mismatches),
                        "mismatches": tick_mismatches,
                    },
                ),
            )

    return checked_count, None


def _load_pairs(
    *,
    expected_trace: TraceReader,
    actual_trace: TraceReader,
    tick_start: int | None,
    tick_end: int | None,
) -> list[_TickPair]:
    expected_rows = {row.tick_index: row for row in expected_trace.iter_ticks(tick_start=tick_start, tick_end=tick_end)}
    actual_rows = {row.tick_index: row for row in actual_trace.iter_ticks(tick_start=tick_start, tick_end=tick_end)}
    all_ticks = sorted(set(expected_rows) | set(actual_rows))
    pairs: list[_TickPair] = []
    for tick in all_ticks:
        exp_row = expected_rows.get(tick)
        act_row = actual_rows.get(tick)
        pairs.append(
            _TickPair(
                tick_index=tick,
                expected_row=exp_row,
                actual_row=act_row,
            ),
        )
    return pairs


def diff_traces(
    *,
    expected_trace_path: Path,
    actual_trace_path: Path,
    tick_start: int | None = None,
    tick_end: int | None = None,
) -> TraceDiffReport:
    with TraceReader(Path(expected_trace_path)) as expected_trace, TraceReader(Path(actual_trace_path)) as actual_trace:
        expected_channels = {str(channel) for channel in expected_trace.meta.channels}
        actual_channels = {str(channel) for channel in actual_trace.meta.channels}
        for required_channel in TRACE_REQUIRED_CHANNELS:
            channel_name = str(required_channel)
            if channel_name not in expected_channels or channel_name not in actual_channels:
                return TraceDiffReport(
                    ok=False,
                    checked_count=0,
                    tick_start=tick_start,
                    tick_end=tick_end,
                    mismatch=TraceMismatch(
                        kind="missing_channel",
                        tick_index=-1,
                        detail={
                            "channel": channel_name,
                            "expected_has": channel_name in expected_channels,
                            "actual_has": channel_name in actual_channels,
                        },
                    ),
                )
        pairs = _load_pairs(
            expected_trace=expected_trace,
            actual_trace=actual_trace,
            tick_start=tick_start,
            tick_end=tick_end,
        )
        try:
            checked_count, mismatch = _first_mismatch(
                pairs=pairs,
                tick_end=tick_end,
            )
        except TraceError as exc:
            raise ValueError(str(exc)) from exc
        return TraceDiffReport(
            ok=(mismatch is None),
            checked_count=checked_count,
            tick_start=tick_start,
            tick_end=tick_end,
            mismatch=mismatch,
        )


def bisect_traces(
    *,
    expected_trace_path: Path,
    actual_trace_path: Path,
    tick_start: int | None = None,
    tick_end: int | None = None,
    window_before: int = 12,
    window_after: int = 6,
) -> TraceBisectReport:
    with TraceReader(Path(expected_trace_path)) as expected_trace, TraceReader(Path(actual_trace_path)) as actual_trace:
        pairs = _load_pairs(
            expected_trace=expected_trace,
            actual_trace=actual_trace,
            tick_start=tick_start,
            tick_end=tick_end,
        )
        if not pairs:
            return TraceBisectReport(
                ok=True,
                first_bad_tick=None,
                checked_count=0,
                mismatch=None,
                window_start=None,
                window_end=None,
            )

        end_tick_bound = pairs[-1].tick_index if tick_end is None else tick_end
        try:
            checked_count, mismatch = _first_mismatch(
                pairs=pairs,
                tick_end=end_tick_bound,
            )
        except TraceError as exc:
            raise ValueError(str(exc)) from exc
        if mismatch is None:
            return TraceBisectReport(
                ok=True,
                first_bad_tick=None,
                checked_count=checked_count,
                mismatch=None,
                window_start=None,
                window_end=None,
            )

        first_bad = mismatch.tick_index
        final_mismatch = mismatch
        left = first_bad - max(0, window_before)
        right = first_bad + max(0, window_after)

        return TraceBisectReport(
            ok=False,
            first_bad_tick=first_bad,
            checked_count=checked_count,
            mismatch=final_mismatch,
            window_start=left,
            window_end=right,
        )


def mismatch_to_json(mismatch: TraceMismatch | None) -> BuiltinObject | None:
    if mismatch is None:
        return None
    return to_builtin_object(
        {
            "kind": mismatch.kind,
            "tick_index": mismatch.tick_index,
            "detail": (None if mismatch.detail is None else to_builtin_object(mismatch.detail, field="mismatch.detail")),
        },
        field="mismatch_json",
    )


def diff_report_to_json(report: TraceDiffReport) -> BuiltinObject:
    return to_builtin_object(
        {
            "schema_version": 1,
            "status": ("ok" if report.ok else "diverged"),
            "checked_count": report.checked_count,
            "tick_start": report.tick_start,
            "tick_end": report.tick_end,
            "mismatch": mismatch_to_json(report.mismatch),
        },
        field="diff_report",
    )


def bisect_report_to_json(report: TraceBisectReport) -> BuiltinObject:
    return to_builtin_object(
        {
            "schema_version": 1,
            "status": ("ok" if report.ok else "diverged"),
            "checked_count": report.checked_count,
            "first_bad_tick": report.first_bad_tick,
            "mismatch": mismatch_to_json(report.mismatch),
            "window_start": report.window_start,
            "window_end": report.window_end,
        },
        field="bisect_report",
    )



from __future__ import annotations

from collections.abc import Callable

from grim.geom import Vec2
from grim.rand import CrandLike

from ...bonuses.fire_bullets import queue_large_hit_decal_streak
from ...effects import FxQueue
from ...projectiles.types import ProjectileHit, ProjectileTemplateId


def queue_projectile_large_streak_decal(
    *,
    hit: ProjectileHit,
    base_angle: float,
    fx_queue: FxQueue,
    rng: CrandLike,
    freeze_origin: Vec2 | None = None,
    spawn_freeze_shard: Callable[[Vec2, float], None] | None = None,
) -> bool:
    type_id = hit.type_id
    if type_id not in (ProjectileTemplateId.GAUSS_GUN, ProjectileTemplateId.FIRE_BULLETS):
        return False
    queue_large_hit_decal_streak(
        hit=hit,
        base_angle=float(base_angle),
        fx_queue=fx_queue,
        rng=rng,
        freeze_origin=freeze_origin,
        spawn_freeze_shard=spawn_freeze_shard,
    )
    return True



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

import msgspec

from grim.assets import RuntimeResources, TextureId, runtime_resources_for
from grim.config import CrimsonConfig
from grim.geom import Vec2
from grim.raylib_api import rl
from grim.terrain_render import GroundRenderer

from ..creatures.anim import creature_corpse_frame_for_type
from ..creatures.runtime import CreaturePool
from ..gameplay import GameplayState
from ..render.frame import RenderFrame
from ..render.rtx.mode import RtxRenderMode
from ..render.terrain_fx import FxQueueTextures, bake_terrain_fx_batch
from ..sim.state_types import PlayerState
from ..sim.terrain_fx import TerrainFxBatch


class RenderResources(msgspec.Struct):
    assets_dir: Path
    world_size: float = 1024.0
    config: CrimsonConfig | None = None

    ground: GroundRenderer | None = None
    fx_textures: FxQueueTextures | None = None
    _pending_terrain_fx_batches: list[TerrainFxBatch] = msgspec.field(default_factory=list)
    _resources: RuntimeResources | None = None

    @property
    def resources(self) -> RuntimeResources:
        resources = self._resources
        assert resources is not None, "runtime resources must be loaded before use"
        return resources

    @resources.setter
    def resources(self, value: RuntimeResources | None) -> None:
        self._resources = value

    def registry_texture(self, texture_id: TextureId) -> rl.Texture:
        resources = self._resources
        if resources is not None:
            return resources.texture(texture_id)
        return runtime_resources_for(self.assets_dir).texture(texture_id)

    def sync_ground_settings(self) -> None:
        if self.ground is None:
            return
        if self.config is None:
            self.ground.texture_scale = 1.0
            return
        self.ground.texture_scale = self.config.display.texture_scale

    def set_ground_textures(
        self,
        *,
        base: rl.Texture,
        overlay: rl.Texture,
        detail: rl.Texture,
    ) -> None:
        self.clear_pending_terrain_fx()
        if self.ground is None:
            self.ground = GroundRenderer(
                texture=base,
                overlay=overlay,
                overlay_detail=detail,
                width=int(self.world_size),
                height=int(self.world_size),
                texture_scale=1.0,
            )
        else:
            self.ground.texture = base
            self.ground.overlay = overlay
            self.ground.overlay_detail = detail
        self.sync_ground_settings()

    def schedule_ground_generation(self, *, seed: int) -> None:
        if self.ground is None:
            return
        self.ground.schedule_generate(seed=seed)

    def process_ground_pending(self) -> None:
        if self.ground is None:
            return
        self.ground.process_pending()
        if self.ground.texture_failed:
            self.clear_pending_terrain_fx()
            return
        if not self.ground.render_target_ready():
            return
        if self.fx_textures is None or not self._pending_terrain_fx_batches:
            return
        pending = tuple(self._pending_terrain_fx_batches)
        self._pending_terrain_fx_batches.clear()
        for batch in pending:
            self._bake_terrain_fx_batch(
                batch,
                corpse_frame_for_type=creature_corpse_frame_for_type,
            )

    def open(self, *, terrain_seed: int) -> None:
        self.close()
        resources = runtime_resources_for(self.assets_dir)
        self._resources = resources

        base = resources.texture(TextureId.TER_Q1_BASE)
        overlay = resources.texture(TextureId.TER_Q1_OVERLAY)
        self.set_ground_textures(base=base, overlay=overlay, detail=base)
        self.schedule_ground_generation(seed=terrain_seed)
        self.fx_textures = FxQueueTextures(
            particles=resources.texture(TextureId.PARTICLES),
            bodyset=resources.texture(TextureId.BODYSET),
        )

    def close(self) -> None:
        if self.ground is not None and self.ground.render_target is not None:
            rl.unload_render_texture(self.ground.render_target)
            self.ground.render_target = None
        self.ground = None

        self._resources = None
        self.fx_textures = None
        self.clear_pending_terrain_fx()

    def clear_pending_terrain_fx(self) -> None:
        self._pending_terrain_fx_batches.clear()

    def _bake_terrain_fx_batch(
        self,
        batch: TerrainFxBatch,
        *,
        corpse_frame_for_type: Callable[[int], int] = creature_corpse_frame_for_type,
    ) -> None:
        if self.ground is None or self.fx_textures is None:
            return
        if batch.is_empty():
            return
        bake_terrain_fx_batch(
            self.ground,
            batch=batch,
            textures=self.fx_textures,
            corpse_frame_for_type=corpse_frame_for_type,
        )

    def consume_terrain_fx_batch(
        self,
        batch: TerrainFxBatch,
        *,
        corpse_frame_for_type: Callable[[int], int] = creature_corpse_frame_for_type,
    ) -> None:
        if batch.is_empty():
            return
        ground = self.ground
        if ground is None or ground.texture_failed or self.fx_textures is None:
            return
        if ground.render_target_ready():
            self._bake_terrain_fx_batch(
                batch,
                corpse_frame_for_type=corpse_frame_for_type,
            )
            return
        self._pending_terrain_fx_batches.append(batch)

    def build_render_frame(
        self,
        *,
        state: GameplayState,
        players: list[PlayerState],
        creatures: CreaturePool,
        camera: Vec2,
        demo_mode_active: bool,
        elapsed_ms: float,
        bonus_anim_phase: float,
        lan_player_rings_enabled: bool,
        lan_local_aim_indicators_only: bool,
        lan_local_player_slot_index: int,
        rtx_mode: RtxRenderMode,
    ) -> RenderFrame:
        return RenderFrame(
            world_size=float(self.world_size),
            demo_mode_active=bool(demo_mode_active),
            config=self.config,
            camera=camera,
            ground=self.ground,
            state=state,
            players=players,
            creatures=creatures,
            resources=self.resources,
            elapsed_ms=float(elapsed_ms),
            bonus_anim_phase=float(bonus_anim_phase),
            lan_player_rings_enabled=bool(lan_player_rings_enabled),
            lan_local_aim_indicators_only=bool(lan_local_aim_indicators_only),
            lan_local_player_slot_index=int(lan_local_player_slot_index),
            rtx_mode=rtx_mode,
        )



from __future__ import annotations

import time
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Protocol


class RenderProfileSink(Protocol):
    def on_pass_duration(self, pass_name: str, duration_ms: float) -> None: ...


_ACTIVE_SINK: RenderProfileSink | None = None
_PASS_STACK: list[tuple[str, int]] = []


def set_active_sink(sink: RenderProfileSink | None) -> RenderProfileSink | None:
    global _ACTIVE_SINK
    prev = _ACTIVE_SINK
    _ACTIVE_SINK = sink
    return prev


def clear_pass_stack() -> None:
    _PASS_STACK.clear()


def current_pass_name() -> str | None:
    if not _PASS_STACK:
        return None
    return str(_PASS_STACK[-1][0])


def begin_pass(name: str) -> None:
    _PASS_STACK.append((str(name), int(time.perf_counter_ns())))


def end_pass(name: str) -> None:
    _ = name
    if not _PASS_STACK:
        return
    pass_name, start_ns = _PASS_STACK.pop()
    duration_ns = max(0, int(time.perf_counter_ns()) - int(start_ns))
    sink = _ACTIVE_SINK
    if sink is None:
        return
    sink.on_pass_duration(str(pass_name), float(duration_ns) / 1_000_000.0)


@contextmanager
def profile_pass(name: str) -> Iterator[None]:
    begin_pass(name)
    try:
        yield
    finally:
        end_pass(name)


__all__ = [
    "RenderProfileSink",
    "begin_pass",
    "clear_pass_stack",
    "current_pass_name",
    "end_pass",
    "profile_pass",
    "set_active_sink",
]



from __future__ import annotations

import math

from grim.color import RGBA
from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rd, rl

from ...projectiles.types import ProjectileTemplateId
from ...sim.world_defs import KNOWN_PROJ_FRAMES
from .common import proj_origin
from .types import ProjectileDrawCtx


def _begin_darken_src_zero_blend() -> None:
    # Native projectile_render switches Plague Spreader to D3D8 SRC=ZERO / DST=INVSRCALPHA.
    rl.rl_set_blend_factors_separate(
        rd.RL_ZERO,
        rd.RL_ONE_MINUS_SRC_ALPHA,
        rd.RL_ZERO,
        rd.RL_ONE,
        rd.RL_FUNC_ADD,
        rd.RL_FUNC_ADD,
    )
    rl.begin_blend_mode(rl.BlendMode.BLEND_CUSTOM_SEPARATE)
    rl.rl_set_blend_factors_separate(
        rd.RL_ZERO,
        rd.RL_ONE_MINUS_SRC_ALPHA,
        rd.RL_ZERO,
        rd.RL_ONE,
        rd.RL_FUNC_ADD,
        rd.RL_FUNC_ADD,
    )


def draw_pulse_gun(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    if ctx.type_id != ProjectileTemplateId.PULSE_GUN:
        return False
    if ctx.texture is None:
        return False

    mapping = KNOWN_PROJ_FRAMES.get(ctx.type_id)
    if mapping is None:
        return True
    grid, frame = mapping
    cell_w = float(ctx.texture.width) / float(grid)

    alpha = float(ctx.alpha)
    life = float(ctx.life)
    if life >= 0.4:
        origin = proj_origin(ctx.proj, ctx.pos)
        dist = origin.distance_to(ctx.pos)

        desired_size = dist * 0.16 * ctx.scale
        if desired_size <= 1e-3:
            return True
        sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
        if sprite_scale <= 1e-6:
            return True

        tint = RGBA(0.1, 0.6, 0.2, alpha * 0.7).to_rl()
        rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
        renderer._draw_atlas_sprite(
            ctx.texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=sprite_scale,
            rotation_rad=ctx.angle,
            tint=tint,
        )
        rl.end_blend_mode()
        return True

    fade = clamp(life * 2.5, 0.0, 1.0)
    fade_alpha = fade * alpha
    if fade_alpha <= 1e-3:
        return True

    desired_size = 56.0 * ctx.scale
    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    tint = RGBA(1.0, 1.0, 1.0, fade_alpha).to_rl()
    rl.begin_blend_mode(rl.BlendMode.BLEND_ADDITIVE)
    renderer._draw_atlas_sprite(
        ctx.texture,
        grid=grid,
        frame=frame,
        pos=ctx.screen_pos,
        scale=sprite_scale,
        rotation_rad=ctx.angle,
        tint=tint,
    )
    rl.end_blend_mode()
    return True


def draw_splitter_or_blade(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    type_id = ctx.type_id
    if type_id not in (ProjectileTemplateId.SPLITTER_GUN, ProjectileTemplateId.BLADE_GUN):
        return False
    if ctx.texture is None:
        return False

    mapping = KNOWN_PROJ_FRAMES.get(type_id)
    if mapping is None:
        return True
    grid, frame = mapping
    cell_w = float(ctx.texture.width) / float(grid)

    if float(ctx.life) < 0.4:
        return True

    origin = proj_origin(ctx.proj, ctx.pos)
    dist = origin.distance_to(ctx.pos)

    desired_size = min(dist, 20.0) * ctx.scale
    if desired_size <= 1e-3:
        return True

    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    rotation_rad = ctx.angle
    rgb = (1.0, 1.0, 1.0)
    if type_id == ProjectileTemplateId.BLADE_GUN:
        rotation_rad = float(int(ctx.proj_index)) * 0.1 - float(renderer.frame.elapsed_ms) * 0.1
        rgb = (0.8, 0.8, 0.8)

    tint = RGBA(rgb[0], rgb[1], rgb[2], float(ctx.alpha)).to_rl()
    renderer._draw_atlas_sprite(
        ctx.texture,
        grid=grid,
        frame=frame,
        pos=ctx.screen_pos,
        scale=sprite_scale,
        rotation_rad=rotation_rad,
        tint=tint,
    )
    return True


def draw_plague_spreader(ctx: ProjectileDrawCtx) -> bool:
    renderer = ctx.renderer
    if ctx.type_id != ProjectileTemplateId.PLAGUE_SPREADER:
        return False
    texture = ctx.texture
    if texture is None:
        return False

    grid = 4
    frame = 2
    cell_w = float(texture.width) / float(grid)

    alpha = float(ctx.alpha)
    life = float(ctx.life)
    if life >= 0.4:
        tint = RGBA(1.0, 1.0, 1.0, alpha).to_rl()

        def draw_plague_quad(*, pos: Vec2, size: float) -> None:
            size = float(size)
            if size <= 1e-3:
                return
            desired_size = size * ctx.scale
            sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
            if sprite_scale <= 1e-6:
                return
            pos_screen = renderer.world_to_screen(pos)
            renderer._draw_atlas_sprite(
                texture,
                grid=grid,
                frame=frame,
                pos=pos_screen,
                scale=sprite_scale,
                rotation_rad=0.0,
                tint=tint,
            )

        _begin_darken_src_zero_blend()
        try:
            draw_plague_quad(pos=ctx.pos, size=60.0)

            offset = Vec2.from_heading(ctx.angle + math.pi) * 15.0
            draw_plague_quad(
                pos=ctx.pos + offset,
                size=60.0,
            )

            phase = float(int(ctx.proj_index)) + float(renderer.frame.elapsed_ms) * 0.01
            cos_phase = math.cos(phase)
            sin_phase = math.sin(phase)
            draw_plague_quad(
                pos=ctx.pos.offset(dx=cos_phase * cos_phase - 5.0, dy=sin_phase * 11.0 - 5.0),
                size=52.0,
            )

            phase_120 = phase + 2.0943952
            sin_phase_120 = math.sin(phase_120)
            draw_plague_quad(
                pos=ctx.pos + Vec2.from_polar(phase_120, 10.0),
                size=62.0,
            )

            phase_240 = phase + 4.1887903
            draw_plague_quad(
                pos=ctx.pos + Vec2(math.cos(phase_240) * 10.0, math.sin(phase_240) * sin_phase_120),
                size=62.0,
            )
        finally:
            rl.end_blend_mode()
        return True

    fade = clamp(life * 2.5, 0.0, 1.0)
    fade_alpha = fade * alpha
    if fade_alpha <= 1e-3:
        return True

    desired_size = (fade * 40.0 + 32.0) * ctx.scale
    sprite_scale = desired_size / cell_w if cell_w > 1e-6 else 0.0
    if sprite_scale <= 1e-6:
        return True

    tint = RGBA(1.0, 1.0, 1.0, fade_alpha).to_rl()
    _begin_darken_src_zero_blend()
    try:
        renderer._draw_atlas_sprite(
            texture,
            grid=grid,
            frame=frame,
            pos=ctx.screen_pos,
            scale=sprite_scale,
            rotation_rad=0.0,
            tint=tint,
        )
    finally:
        rl.end_blend_mode()
    return True


__all__ = ["draw_plague_spreader", "draw_pulse_gun", "draw_splitter_or_blade"]



from __future__ import annotations

"""Presentation-phase deterministic helpers."""

from .projectile_decals import queue_projectile_large_streak_decal

__all__ = ["queue_projectile_large_streak_decal"]
